import random
from uuid import uuid4


def get_uuid():
    return uuid4().hex


def get_id(prefix, length):
    return prefix + "".join(random.choices(uuid4().hex, k=length)).upper()


def get_base_id():
    return get_id("", length=4)


# Types of pechas


def get_initial_pecha_id():
    return get_id(prefix="I", length=8)


def get_open_pecha_id():
    return get_id(prefix="O", length=8)


def get_diplomatic_id():
    return get_id(prefix="D", length=8)


def get_work_id():
    return get_id(prefix="W", length=8)


def get_alignment_id():
    return get_id(prefix="A", length=8)


def get_collection_id():
    return get_id(prefix="C", length=8)



import re
from pathlib import Path

from openpecha.blupdate import *
from openpecha.utils import download_pecha, dump_yaml, load_yaml


def get_meta(pecha_id, branch="main", pecha_path=None):
    """Return meta data of pecha

    Args:
        pecha_id (str): pecha id
        pecha_path (path, optional): pecha path. Defaults to None.

    Returns:
        dict: meta data of pecha
    """
    if not pecha_path:
        pecha_path = download_pecha(pecha_id, branch=branch)
    meta = load_yaml(Path(f"{pecha_path}/{pecha_id}.opf/meta.yml"))
    return meta


def get_pagination_layer(pecha_id, vol_meta, branch="main", pecha_path=None):
    """Return pagination layer of a volume from pecha opf of mentioned pecha id

    Args:
        pecha_id (str): pecha id
        vol_meta (dict): volume meta data
        pecha_path (path, optional): pecha path. Defaults to None.

    Returns:
        dict: pagination layer of the volume passed in parameter from mentioned pecha_id in parameter
    """
    vol_id = re.sub(r"\..+", "", vol_meta["base_file"])
    if not pecha_path:
        pecha_path = download_pecha(pecha_id, branch=branch)
    pagination_layer = load_yaml(
        Path(f"{pecha_path}/{pecha_id}.opf/layers/{vol_id}/Pagination.yml")
    )
    return pagination_layer


def get_vol_info(pecha_id, branch="main", pecha_path=None):
    """Return volumes information in pecha
    Args:
        pecha_id (str): pecha id
        pecha_path (path, optional): pecha path. Defaults to None.

    Returns:
        list: list of dict containing each volume's detail
    """
    vols = []
    meta_data = get_meta(pecha_id, branch=branch, pecha_path=pecha_path)
    for vol_uuid, vol_info in meta_data["source_metadata"]["volumes"].items():
        cur_vol = {
            "id": vol_uuid,
            "title": vol_info["title"],
            "image_group_id": vol_info["image_group_id"],
        }
        vols.append(cur_vol)
    return vols


def get_pages_info(pecha_id, vol_id, branch="main", pecha_path=None):
    """Return page annotation's uuids in mentioned volume number

    Args:
        pecha_id (str): pecha id
        vol_id (str): volume uuid
        pecha_path (path, optional): pecha path. Defaults to None.

    Returns:
        list: page annotation's uuids in mentioned volume number in sorted order
    """

    pages = []
    pecha_meta = get_meta(pecha_id, branch=branch, pecha_path=pecha_path)
    vol_meta = pecha_meta["source_metadata"]["volumes"][vol_id]
    pagination_layer = get_pagination_layer(
        pecha_id, vol_meta, branch=branch, pecha_path=pecha_path
    )
    pages = [ann_uuid for ann_uuid, ann in pagination_layer["annotations"].items()]
    return pages


def get_base_text(pecha_id, vol_meta, branch="main", pecha_path=None):
    """Return base text of the mentioned volume number

    Args:
        pecha_id (str): pecha id
        vol_meta (dict): volume meta data
        pecha_path (path, optional): pecha path. Defaults to None.

    Returns:
        str: base text of volume
    """
    base_text = ""
    base_fn = vol_meta["base_file"]
    if not pecha_path:
        pecha_path = download_pecha(pecha_id, branch=branch)
    base_text = Path(f"{pecha_path}/{pecha_id}.opf/base/{base_fn}").read_text(
        encoding="utf-8"
    )
    return base_text


def get_page_content(page_ann, base_text):
    """Return page content from base text

    Args:
        page_ann (dict): span of page, page reference, page index are included
        base_text (str): base text

    Returns:
        str: page content
    """
    page_span = page_ann["span"]
    page_start_idx = page_span["start"]
    page_end_idx = page_span["end"]
    page_content = base_text[page_start_idx : page_end_idx + 1]
    return page_content


def get_page_image_url(page_ann, vol_meta):
    """Return image url of the page

    Args:
        page_ann (dict): page annotation
        vol_meta (dict): volume meta data

    Returns:
        str: image url of page
    """
    cur_image_grp_id = ""
    cur_image_grp_id = vol_meta["image_group_id"]
    image_ref = page_ann["reference"]
    image_url = f"https://iiif.bdrc.io/bdr:{cur_image_grp_id}::{image_ref}/full/max/0/default.jpg"
    return image_url


def get_page(pecha_id, vol_id, page_id, branch="main", pecha_path=None):
    """Return page detail of mentioned page id

    Args:
        pecha_id (strt): pecha id
        vol_id (str): volume uuid
        page_id (uuid): uuid of page
        pecha_path (path, optional): pecha path. Defaults to None.

    Returns:
        dict: page content and image url of page will be return
    """
    page_info = {"content": None, "image_url": None}
    pecha_meta = get_meta(pecha_id, branch=branch, pecha_path=pecha_path)
    vol_meta = pecha_meta["source_metadata"]["volumes"][vol_id]
    pagination_layer = get_pagination_layer(
        pecha_id, vol_meta, branch=branch, pecha_path=pecha_path
    )
    base_text = get_base_text(pecha_id, vol_meta, branch=branch, pecha_path=pecha_path)
    cur_page_ann = pagination_layer["annotations"][page_id]
    page_info["content"] = get_page_content(cur_page_ann, base_text)
    page_info["image_url"] = get_page_image_url(cur_page_ann, vol_meta)
    return page_info


def get_new_vol(old_vol, old_page, new_page_content):
    """Return new volume content by replacing old page with new page in old volume

    Args:
        old_vol (str): old base text
        old_page (str): old page content
        new_page_content (str): new page content

    Returns:
        str: new base text
    """
    old_page = old_page.strip()
    new_page = new_page_content.strip()
    new_vol = old_vol.replace(old_page, new_page)
    return new_vol


def update_layer(pecha_path, pecha_id, vol_id, old_layers, updater):
    """Save updated layer using updater object

    Args:
        pecha_path (path): pecha path
        pecha_id (str): pecha id
        vol_id (str): volume id
        old_layers (dict): contains layer name and layer details
        updater (obj): BaseLayerUpdate object
    """
    for layer_name, old_layer in old_layers.items():
        update_ann_layer(old_layer, updater)
        new_layer_path = Path(
            f"{pecha_path}/{pecha_id}.opf/layers/{vol_id}/{layer_name}.yml"
        )
        dump_yaml(old_layer, new_layer_path)
        print(f"INFO: {vol_id} {layer_name} has been updated...")


def get_old_layers(pecha_path, pecha_id, vol_id):
    """Return old layer of the mentioned volume id

    Args:
        pecha_path (path): pecha path
        pecha_id (str): pecha id
        vol_id (str): volume id

    Returns:
        dict: layer name as key and its detail as value
    """
    old_layers = {}
    layer_paths = list(Path(f"{pecha_path}/{pecha_id}.opf/layers/{vol_id}").iterdir())
    for layer_path in layer_paths:
        layer_name = layer_path.stem
        layer_content = load_yaml(layer_path)
        old_layers[layer_name] = layer_content
    return old_layers


def update_old_layers(pecha_path, pecha_id, old_vol, new_vol, vol_meta):
    """Update old layer while change recorded in base text

    Args:
        pecha_path (path): pecha path
        pecha_id (str): pecah id
        old_vol (str): old base text
        new_vol (str): new base text
        vol_meta (dict): volume meta data
    """
    vol_id = re.sub(r"\..+", "", vol_meta["base_file"])
    updater = Blupdate(old_vol, new_vol)
    old_layers = get_old_layers(pecha_path, pecha_id, vol_id)
    update_layer(pecha_path, pecha_id, vol_id, old_layers, updater)


def update_base(pecha_path, pecha_id, vol_meta, new_vol):
    """Update old base text by overwriting the new base text

    Args:
        pecha_path (path): pecha path
        pecha_id (str): pecha id
        vol_meta (dict): volume meta data
        new_vol (str): updated base text
    """
    base_fn = vol_meta["base_file"]
    Path(f"{pecha_path}/{pecha_id}.opf/base/{base_fn}").write_text(
        new_vol, encoding="utf-8"
    )
    print(f"INFO: {base_fn} base updated..")


def update_sub_text(old_pecha_idx, text_info, vol_num, vol_offset, page_start):
    """Update span of sub text acording to offset

    Args:
        old_pecha_idx (dict): old pecha index
        text_info (list): contains text uuid and text annotation details
        vol_num (int): volume number
        vol_offset (int): offset between new and old base text
        page_start (int): starting span of the page which you want to save

    Returns:
        dict: old pecha with updated sub text span
    """
    new_pecha_idx = old_pecha_idx
    check_next_sub_text = True
    text_uuid = text_info[0]
    text_ann = text_info[1]
    for sub_text_uuid, sub_text_ann in text_ann["parts"].items():
        sub_text_span = sub_text_ann["span"]
        for vol_walker, vol_span in enumerate(sub_text_span):
            if vol_span["vol"] == vol_num and vol_span["start"] > page_start:
                new_pecha_idx["annotations"][text_uuid]["parts"][sub_text_uuid]["span"][
                    vol_walker
                ]["start"] += vol_offset
            if vol_span["vol"] == vol_num and vol_span["end"] >= page_start:
                new_pecha_idx["annotations"][text_uuid]["parts"][sub_text_uuid]["span"][
                    vol_walker
                ]["end"] += vol_offset
            elif vol_span["vol"] > vol_num:
                check_next_sub_text = False
                break
        if not check_next_sub_text:
            break
    return new_pecha_idx


def update_index(vol_offset, vol_meta, page_start, old_pecha_idx):
    """Return updated index

    Args:
        vol_offset (int): length difference between old base text and new base text
        vol_meta (dict): volume meta data
        page_start (int): starting index of page that has been affected
        old_pecha_idx (dict): old pecha index

    Returns:
        dict: new index with new span using offset to adjust
    """
    new_pecha_idx = old_pecha_idx
    vol_num = int(re.search(r"\d+", vol_meta["base_file"])[0])
    check_next_text = True
    if vol_offset != 0:
        for text_uuid, text_ann in new_pecha_idx["annotations"].items():
            if text_ann["parts"]:
                text_info = [text_uuid, text_ann]
                new_pecha_idx = update_sub_text(
                    new_pecha_idx, text_info, vol_num, vol_offset, page_start
                )
            text_span = text_ann["span"]
            for vol_walker, vol_span in enumerate(text_span):
                if vol_span["vol"] == vol_num and vol_span["start"] > page_start:
                    new_pecha_idx["annotations"][text_uuid]["span"][vol_walker][
                        "start"
                    ] += vol_offset
                if vol_span["vol"] == vol_num and vol_span["end"] >= page_start:
                    new_pecha_idx["annotations"][text_uuid]["span"][vol_walker][
                        "end"
                    ] += vol_offset
                elif vol_span["vol"] > vol_num:
                    check_next_text = False
                    break
            if not check_next_text:
                break
    return new_pecha_idx


def save_page(pecha_id, vol_id, page_id, page_content, branch="main", pecha_path=None):
    """Save new page in opf

    Args:
        pecha_id (str): pecha id
        vol_id (str): volume uuid
        page_id (str): uuid of the page that needs to be saved
        page_content (str): new page content that needs to be saved
        pecha_path (path, optional): pecha path. Defaults to None.

    Returns:
        path: pecha path of updated opf
    """
    pecha_meta = get_meta(pecha_id, branch=branch, pecha_path=pecha_path)
    vol_meta = pecha_meta["source_metadata"]["volumes"][vol_id]
    if not pecha_path:
        pecha_path = download_pecha(pecha_id, branch=branch)
    old_pecha_idx = load_yaml(Path(f"{pecha_path}/{pecha_id}.opf/index.yml"))
    pagination_layer = get_pagination_layer(
        pecha_id, vol_meta, branch=branch, pecha_path=pecha_path
    )
    old_vol = get_base_text(pecha_id, vol_meta, branch=branch, pecha_path=pecha_path)
    cur_page_ann = pagination_layer["annotations"][page_id]
    old_page = get_page_content(cur_page_ann, old_vol)
    new_vol = get_new_vol(old_vol, old_page, new_page_content=page_content)
    vol_offset = len(new_vol) - len(old_vol)
    new_pecha_idx = update_index(
        vol_offset, vol_meta, cur_page_ann["span"]["start"], old_pecha_idx
    )
    update_old_layers(pecha_path, pecha_id, old_vol, new_vol, vol_meta)
    update_base(pecha_path, pecha_id, vol_meta, new_vol)
    new_pecha_idx_path = dump_yaml(
        new_pecha_idx, Path(f"{pecha_path}/{pecha_id}.opf/index.yml")
    )
    return pecha_path



"""
Formatter for HFML annotations in the text

This module implements all classes necessary to format HFML annotation to OpenPecha format.
HFML (Human Friendly Markup Language) contains tagset used for structuring and annotating the text.
"""
import re
from json import encoder
from pathlib import Path

import yaml

from ..utils import Vol2FnManager, dump_yaml, load_yaml
from .formatter import BaseFormatter
from .layers import *
from .layers import AnnType, _attr_names


class HFMLFormatter(BaseFormatter):
    """
    OpenPecha Formatter for digitized wooden-printed Pecha based on annotation scheme from Esukhia.
    """

    def __init__(self, output_path=None, metadata=None, is_book=False):
        super().__init__(output_path, metadata)
        self.is_book = is_book
        self.base_text = ""
        self.vol_walker = 0
        self.book_title = []
        self.book_number = []
        self.poti_title = []
        self.author = []
        self.chapter_title = []
        self.topic_id = []  # made class variable as it needs to update cross poti
        self.current_topic_id = (
            []
        )  # made class variable as it needs to update cross poti
        self.sub_topic = []  # made class variable as it needs to update cross poti
        self.page = []
        self.error_id = []
        self.archaic_word_id = []
        self.abs_er_id = []
        self.notes_id = []
        self.sub_topic_Id = []  # made class variable as it needs to update cross poti
        self.topic_info = []
        self.topic_local_id = []
        self.sub_topic_info = []
        self.sub_topic_local_id = []
        self.cur_sub = []
        self.author_pattern = []
        self.citation_pattern = []
        self.sabche_pattern = []
        self.tsawa_pattern = []
        self.yigchung_pattern = []
        self.durchen_pattern = []

    def text_preprocess(self, text):
        if text[0] == "\ufeff":
            text = text[1:]

        if not self.is_book:
            return text

        p = r"\[p\]"
        lines = text.splitlines()
        result_text = ""
        para = False
        for line in lines:
            if re.search(p, line):
                if para:
                    para = False
                    result_text += "\n"
                else:
                    para = True
            elif re.search(p, line) is None:
                if para:
                    result_text += line
                else:
                    result_text += line + "\n"
        return result_text

    def _load_metadata(self):
        if self.metadata:
            return self.metadata
        meta_fn = self.dirs["opf_path"] / "meta.yml"
        if meta_fn.is_file():
            return load_yaml(meta_fn)
        else:
            return {}

    def _save_metadata(self, **kwargs):
        meta_fn = self.dirs["opf_path"] / "meta.yml"
        if kwargs:
            self.metadata.update(kwargs)
        if "id" not in self.metadata:
            self.metadata["id"] = f"opecha:{self.pecha_path.name}"
        dump_yaml(self.metadata, meta_fn)

    def get_input(self, input_path):
        fns = list(input_path.iterdir())
        fns_len = len(fns)
        for fn in sorted(fns):
            yield self.text_preprocess(fn.read_text()), fn.name, fns_len

    def total_pattern(self, pat_list, annotated_line):
        """ It calculates the length of all the annotation detected in a line.

        Args:
            pat_list (dict): It contains all the annotation's regex pattern as value and name of annotation as key.
            annotated_line (str): It contains the annotated line to be process.

        Return:
            total_length (int): It accumulates as annotations are detected in the line.
        """
        total_length = 0  # total length of annotation detected in a line
        for pattern in [
            "line_pattern",
            "topic_pattern",
            "sub_topic_pattern",
            "note_pattern",
            "start_cit_pattern",
            "start_sabche_pattern",
            "start_tsawa_pattern",
            "start_yigchung_pattern",
            "start_durchen_pattern",
        ]:
            if re.search(pat_list[pattern], annotated_line):
                match_list = re.finditer(
                    pat_list[pattern], annotated_line
                )  # list of match object of given pattern in line
                for match in match_list:
                    total_length = total_length + len(match[0])

        if re.search(pat_list["error_pattern"], annotated_line):
            errors = re.finditer(
                pat_list["error_pattern"], annotated_line
            )  # list of match object of error pattern in line
            for error in errors:
                if error.group(1):
                    starting_point = 2
                else:
                    starting_point = 1
                error_part = error[0].split(",")[0][starting_point:]
                total_length = total_length + (len(error[0]) - len(error_part))

        if re.search(pat_list["archaic_word_pattern"], annotated_line):
            archaic_words = re.finditer(
                pat_list["archaic_word_pattern"], annotated_line
            )  # list of match object of error pattern in line
            for archaic_word in archaic_words:
                if archaic_word.group(1):
                    starting_point = 2
                else:
                    starting_point = 1
                archaic_part = archaic_word[0].split(",")[0][starting_point:]
                total_length = total_length + (len(archaic_word[0]) - len(archaic_part))

        if re.search(pat_list["abs_er_pattern"], annotated_line):
            abs_ers = re.finditer(
                pat_list["abs_er_pattern"], annotated_line
            )  # list of match of abs_er pattern in line
            for abs_er in abs_ers:
                if abs_er.group(1):
                    total_length += 3
                else:
                    total_length += 2
        for pattern in [
            "author_pattern",
            "book_title_pattern",
            "poti_title_pattern",
            "chapter_title_pattern",
            "book_number_pattern",
        ]:
            title_pattern = re.search(pat_list[pattern], annotated_line)
            if title_pattern:
                if title_pattern.group(1):
                    total_length += 5
                else:
                    total_length += 4

        for pattern in [
            "end_cit_pattern",
            "end_sabche_pattern",
            "end_tsawa_pattern",
            "end_yigchung_pattern",
            "end_durchen_pattern",
        ]:
            end_patterns = re.findall(
                pat_list[pattern], annotated_line
            )  # list of match of citation pattern in line
            total_length = total_length + 2 * len(end_patterns)

        return total_length

    def merge(self, start_list, end_list):
        """ It merges two list.
        The starting  and ending of annotation(citaion,yigchung,sabche and tsawa) are stored in two list.
        Merging these two list will generate a list in which both starting and ending of an annotation together in a tuple.
        It is applicable only if the annotaions are not cross volume.
        Args:
            start_list (list): It contains index of where starting annotations(citaion,yigchung,sabche and tsawa) are detected.
            end_list (list): It contains index of where ending annotations(citaion,yigchung,sabche and tsawa) are detected.

        Return:
            result (list): It contains tuples where starting and ending of an annotation(citaion,yigchung,sabche and tsawa) is combined.
        """
        walker = 0
        result = []
        while walker < len(end_list) and walker < len(start_list):
            result.append(
                (
                    start_list[walker][0],
                    {"span": Span(start_list[walker][1], end_list[walker])},
                )
            )
            walker += 1
        return result

    def search_before(self, ann, pat_list, line):
        """ It calculates the length of annotation detected in a given line before a given annotation.
        Args:
            ann (match object): It is a match object of the annotation of which we want to calculate
                                the length of any annotation detected before it.
            pat_list (dict): It contains all the annotation's regex pattern as value and name of annotation as key.
            line (str): It contains the line in which we wants to calculate the legth of annotation found before the given annotation.
        Return:
            length_before (int): It accumalates as we detect annotation which is before the given annotation and
                                finally gives the total length of annotation caught before the given annotation in the given line.
        """
        length_before = 0
        for pp in [
            "line_pattern",
            "topic_pattern",
            "sub_topic_pattern",
            "note_pattern",
            "start_cit_pattern",
            "start_sabche_pattern",
            "start_tsawa_pattern",
            "start_yigchung_pattern",
            "start_durchen_pattern",
        ]:
            if re.search(pat_list[pp], line):
                match_list = re.finditer(
                    pat_list[pp], line
                )  # list of match object of given pattern in line
                for match in match_list:
                    if ann.start() > match.start():
                        length_before = length_before + len(match[0])

        if re.search(pat_list["error_pattern"], line):
            errors = re.finditer(
                pat_list["error_pattern"], line
            )  # list of match object of error pattern in line
            for error in errors:
                if error.group(1):
                    starting_point = 2
                else:
                    starting_point = 1
                if ann.start() > error.start():
                    error_part = error[0].split(",")[0][starting_point:]
                    length_before = length_before + (len(error[0]) - len(error_part))

        if re.search(pat_list["archaic_word_pattern"], line):
            archaic_words = re.finditer(
                pat_list["archaic_word_pattern"], line
            )  # list of match object of error pattern in line
            for archaic_word in archaic_words:
                if archaic_word.group(1):
                    starting_point = 2
                else:
                    starting_point = 1
                if ann.start() > archaic_word.start():
                    archaic_part = archaic_word[0].split(",")[0][starting_point:]
                    length_before = length_before + (
                        len(archaic_word[0]) - len(archaic_part)
                    )

        if re.search(pat_list["abs_er_pattern"], line):
            abs_ers = re.finditer(
                pat_list["abs_er_pattern"], line
            )  # list of match object of abs_er pattern in line
            for abs_er in abs_ers:
                if ann.start() > abs_er.start():
                    if abs_er.group(1):
                        pat_len = 3
                    else:
                        pat_len = 2
                    length_before = length_before + pat_len

        for pp in [
            "author_pattern",
            "book_title_pattern",
            "poti_title_pattern",
            "chapter_title_pattern",
            "book_number_pattern",
        ]:
            title_pattern = re.search(pat_list[pp], line)
            if title_pattern:
                if ann.start() > title_pattern.start():
                    if title_pattern.group(1):
                        length_before += 4
                    else:
                        length_before += 3

        for pp in [
            "end_cit_pattern",
            "end_sabche_pattern",
            "end_tsawa_pattern",
            "end_yigchung_pattern",
            "end_durchen_pattern",
        ]:
            end_patterns = re.finditer(
                pat_list[pp], line
            )  # list of match object of citation pattern in line
            for end_pattern in end_patterns:
                if ann.start() > end_pattern.start():
                    length_before = length_before + 2

        return length_before

    def base_extract(self, pat_list, annotated_line):
        """ It extract the base text from annotated text.
        Args:
            pat_list (dict): It contains all the annotation's regex pattern as value and name of annotation as key.
            annotated_line (str): It contains the annotated line from which we want to extract the base text.

        Return:
            base_line (str): It contains the base text which is being extracted from the given annotated line.
        """
        base_line = (
            annotated_line  # stores the base_line which is line without annotation
        )
        for pattern in [
            "line_pattern",
            "topic_pattern",
            "sub_topic_pattern",
            "note_pattern",
            "start_cit_pattern",
            "end_cit_pattern",
            "start_sabche_pattern",
            "end_sabche_pattern",
            "start_tsawa_pattern",
            "end_tsawa_pattern",
            "start_yigchung_pattern",
            "end_yigchung_pattern",
            "start_durchen_pattern",
            "end_durchen_pattern",
        ]:
            base_line = re.sub(pat_list[pattern], "", base_line)

        for pattern in [
            "author_pattern",
            "book_title_pattern",
            "poti_title_pattern",
            "chapter_title_pattern",
            "book_number_pattern",
        ]:
            title_pattern = re.search(pat_list[pattern], annotated_line)
            if title_pattern:
                if title_pattern.group(1):
                    starting_point = 4
                else:
                    starting_point = 3
                title = title_pattern[0][starting_point:-1]
                base_line = re.sub(pat_list[pattern], title, base_line, 1)

        if re.search(pat_list["error_pattern"], annotated_line):
            errors = re.finditer(
                pat_list["error_pattern"], annotated_line
            )  # list of match object of error pattern in line
            for error in errors:
                if error.group(1):
                    starting_point = 2
                else:
                    starting_point = 1
                error_part = error[0].split(",")[0][starting_point:]
                base_line = re.sub(pat_list["error_pattern"], error_part, base_line, 1)

        if re.search(pat_list["archaic_word_pattern"], annotated_line):
            archaic_words = re.finditer(
                pat_list["archaic_word_pattern"], annotated_line
            )  # list of match object of error pattern in line
            for archaic_word in archaic_words:
                if archaic_word.group(1):
                    starting_point = 2
                else:
                    starting_point = 1
                archaic_part = archaic_word[0].split(",")[0][starting_point:]
                base_line = re.sub(
                    pat_list["archaic_word_pattern"], archaic_part, base_line, 1
                )

        if re.search(pat_list["abs_er_pattern"], annotated_line):
            abs_ers = re.finditer(
                pat_list["abs_er_pattern"], annotated_line
            )  # list of match object of abs_er pattern in line
            for abs_er in abs_ers:
                if abs_er.group(1):
                    starting_point = 2
                else:
                    starting_point = 1
                base_line = re.sub(
                    pat_list["abs_er_pattern"],
                    abs_er[0][starting_point:-1],
                    base_line,
                    1,
                )

        return base_line

    def get_local_id(self, match_obj):
        if match_obj.group(1):
            return ord(match_obj.group(1))
        else:
            return None

    def parse_start_ann(self, ann, pat_list, walker, line):
        pat_len_before_ann = self.search_before(ann, pat_list, line)
        local_id = self.get_local_id(ann)
        ann_start = ann.start() + walker - pat_len_before_ann
        return (local_id, ann_start)

    def parse_end_ann(self, ann, pat_list, walker, line):
        pat_len_before_ann = self.search_before(ann, pat_list, line)
        ann_end = ann.start() + walker - pat_len_before_ann - 1
        return ann_end

    def parse_payload_ann(self, ann, pat_list, walker, line):
        local_id = self.get_local_id(ann)
        payload = ann[0].split(",")[1][:-1]  # extracting the modern word
        if local_id:
            annotation = ann[0].split(",")[0][2:]  # extracting the error component
        else:
            annotation = ann[0].split(",")[0][1:]
        pat_len_before_ann = self.search_before(ann, pat_list, line)
        start_ann = ann.start() + walker - pat_len_before_ann
        end_ann = start_ann + len(annotation) - 1
        span = Span(start_ann, end_ann)
        return (local_id, span, payload)

    def build_layers(self, m_text, num_vol):

        char_walker = 0  # tracker variable through out the text

        cur_vol_pages = (
            []
        )  # list variable to store page annotation according to base string index eg : [(startPage,endPage)]
        cur_vol_error_id = (
            []
        )  # list variable to store error annotation according to base string index eg : [(es,ee,'suggestion')]
        cur_vol_abs_er_id = []  # list variable to store abs_er annotation
        cur_vol_archaic_id = []
        note_id = []  # list variable to store note annotation '#"
        pg_refs = []  # lsit variable to store page info component
        pg_ann = []  # list variable to store page annotation content
        pg_tid = []

        author_titles = []
        book_titles = []
        book_numbers = []
        poti_titles = []
        chapter_titles = []
        start_cit_patterns = (
            []
        )  # list variable to store index of start citation pattern => (g
        end_cit_patterns = (
            []
        )  # list variable to store index of start citation pattern => g)
        start_sabche_pattern = (
            []
        )  # list variable to store index of start sabche pattern => (q
        end_sabche_pattern = (
            []
        )  # list variable to store index of end sabche pattern => q)
        start_tsawa_pattern = (
            []
        )  # list variable to store index of start tsawa pattern => (m
        end_tsawa_pattern = (
            []
        )  # list variable to store index of end sabche pattern => m)
        start_yigchung_pattern = (
            []
        )  # list variable to store index of start yigchung pattern => (y
        end_yigchung_pattern = (
            []
        )  # list variable to store index of end yigchung pattern => y)
        start_durchen_pattern = (
            []
        )  # list variable to store index of start durchen pattern => <d
        end_durchen_pattern = (
            []
        )  # list variable to store index of end durchen pattern => d>

        pat_list = {
            "author_pattern": r"\<([𰵀-󴉱])?au.+?\>",
            "book_title_pattern": r"\<([𰵀-󴉱])?k1.+?\>",
            "book_number_pattern": r"\<([𰵀-󴉱])?k4.+?\>",
            "poti_title_pattern": r"\<([𰵀-󴉱])?k2.+?\>",
            "chapter_title_pattern": r"\<([𰵀-󴉱])?k3.+?\>",
            "page_pattern": r"〔([𰵀-󴉱])?\d+〕",
            "line_pattern": r"\[\w+\.\d+\]",
            "topic_pattern": r"\{([𰵀-󴉱])?\w+\}",
            "start_cit_pattern": r"\<([𰵀-󴉱])?g",
            "end_cit_pattern": r"g\>",
            "start_sabche_pattern": r"\<([𰵀-󴉱])?q",
            "end_sabche_pattern": r"q\>",
            "start_tsawa_pattern": r"\<([𰵀-󴉱])?m",
            "end_tsawa_pattern": r"m\>",
            "start_yigchung_pattern": r"\<([𰵀-󴉱])?y",
            "end_yigchung_pattern": r"y\>",
            "start_durchen_pattern": r"\<([𰵀-󴉱])?d",
            "end_durchen_pattern": r"d\>",
            "sub_topic_pattern": r"\{([𰵀-󴉱])?\w+\-\w+\}",
            "error_pattern": r"\<([𰵀-󴉱])?\S+?\,\S+?\>",
            "archaic_word_pattern": r"\{([𰵀-󴉱])?\S+?\,\S+?\}",
            "abs_er_pattern": r"\[([𰵀-󴉱])?[^0-9].*?\]",
            "note_pattern": r"#([𰵀-󴉱])?",
        }

        start_page = 0  # starting index of page
        end_page = 0  # ending index of page
        start_line = 0  # starting index of line
        end_line = 0  # ending index of line
        start_title = 0  # starting of the title component
        end_title = 0  # ending of the title component
        start_topic = 0  # starting index of topic_Id
        end_topic = 0  # ending index of topic_Id
        start_sub_topic = 0  # starting index of sub_topic_Id
        end_sub_topic = 0  # ending index of sub_topic_Id
        start_error = 0  # starting index of error
        end_error = 0  # ending index of error
        start_abs_er = 0  # starting index of abs_er
        end_abs_er = 0  # ending index of abs_er
        note = 0  # index of notes

        text_lines = m_text.splitlines()  # list of all the lines in the text
        n_line = len(text_lines)  # number of lines in the text

        for idx, line in enumerate(text_lines):
            line = line.strip()
            start_line = char_walker
            length = len(line)
            pat_len_before_ann = 0  # length of pattern recognised before  annotation
            if re.search(
                pat_list["page_pattern"], line
            ):  # checking current line contains page annotation or not
                start_page = end_page
                end_page = end_line
                local_id = self.get_local_id(re.search(pat_list["page_pattern"], line))
                pg_tid.append(local_id)
                page_reference = line[re.search(pat_list["page_pattern"], line).end() :]
                if local_id:
                    pg_ann.append(re.search(pat_list["page_pattern"], line)[0][2:-1])
                else:
                    pg_ann.append(re.search(pat_list["page_pattern"], line)[0][1:-1])
                pg_refs.append(page_reference)
                if len(pg_refs) >= 2:
                    cur_vol_pages.append(
                        (
                            pg_tid[-2],
                            Page(
                                Span(start_page, end_page),
                                imgnum=pg_ann[-2],
                                page_ref=pg_refs[-2],
                            ),
                        )
                    )
                    if start_page < end_page:  # to ignore the empty pages
                        char_walker = char_walker + 1  # To accumulate the \n character
                        end_page = end_page + 3
                    self.base_text = self.base_text + "\n"
                continue

            for pp in [
                "author_pattern",
                "book_title_pattern",
                "poti_title_pattern",
                "chapter_title_pattern",
                "book_number_pattern",
            ]:
                title_pattern = re.search(pat_list[pp], line)
                if title_pattern:
                    local_id = self.get_local_id(title_pattern)
                    pat_len_before_ann = self.search_before(
                        title_pattern, pat_list, line
                    )
                    start_title = (
                        title_pattern.start() + char_walker - pat_len_before_ann
                    )
                    if local_id:
                        end_title = start_title + len(title_pattern[0]) - 6
                    else:
                        end_title = start_title + len(title_pattern[0]) - 5
                    span = Span(start_title, end_title)
                    if pp == "author_pattern":
                        author_titles.append((local_id, Author(span)))
                    if pp == "book_title_pattern":
                        book_titles.append((local_id, BookTitle(span)))
                    if pp == "book_number_pattern":
                        book_numbers.append((local_id, BookNumber(span)))
                    if pp == "poti_title_pattern":
                        poti_titles.append((local_id, PotiTitle(span)))
                        if local_id:
                            end_topic = len(title_pattern[0][2:])
                            end_sub_topic = len(title_pattern[0][2:])
                        else:
                            end_topic = len(title_pattern[0][1:])
                            end_sub_topic = len(title_pattern[0][1:])
                    if pp == "chapter_title_pattern":
                        chapter_titles.append((local_id, Chapter(span)))

            if re.search(
                pat_list["sub_topic_pattern"], line
            ):  # checking current line contain sub_topicID annotation or not
                sub_topic_match = re.search(pat_list["sub_topic_pattern"], line)
                local_id = self.get_local_id(sub_topic_match)
                if local_id:
                    self.sub_topic_info.append(sub_topic_match[0][2:-1])
                else:
                    self.sub_topic_info.append(sub_topic_match[0][1:-1])
                self.sub_topic_local_id.append(local_id)
                pat_len_before_ann = self.search_before(sub_topic_match, pat_list, line)
                if start_sub_topic == 0:
                    start_sub_topic = end_sub_topic
                    end_sub_topic = (
                        sub_topic_match.start() + char_walker - pat_len_before_ann
                    )

                    if start_sub_topic < end_sub_topic:
                        if len(self.sub_topic_info) >= 2:
                            span = CrossVolSpan(
                                self.vol_walker + 1, start_sub_topic, end_sub_topic
                            )
                            self.sub_topic_Id.append(
                                (
                                    self.sub_topic_local_id[-2],
                                    {"work_id": self.sub_topic_info[-2], "span": span},
                                )
                            )
                            end_sub_topic = end_sub_topic
                        else:
                            span = CrossVolSpan(
                                self.vol_walker + 1, start_sub_topic, end_sub_topic
                            )
                            self.sub_topic_Id.append(
                                (
                                    self.sub_topic_local_id[-1],
                                    {"work_id": self.sub_topic_info[-1], "span": span},
                                )
                            )
                            end_sub_topic = end_sub_topic
                else:
                    start_sub_topic = end_sub_topic
                    end_sub_topic = (
                        sub_topic_match.start() + char_walker - pat_len_before_ann
                    )

                    if start_sub_topic < end_sub_topic:
                        span = CrossVolSpan(
                            self.vol_walker + 1, start_sub_topic, end_sub_topic
                        )
                        self.sub_topic_Id.append(
                            (
                                self.sub_topic_local_id[-2],
                                {"work_id": self.sub_topic_info[-2], "span": span},
                            )
                        )
                        end_sub_topic = end_sub_topic

            if re.search(
                pat_list["topic_pattern"], line
            ):  # checking current line contain topicID annotation or not
                topic = re.search(pat_list["topic_pattern"], line)
                pat_len_before_ann = self.search_before(topic, pat_list, line)
                local_id = self.get_local_id(topic)
                if local_id:
                    self.topic_info.append(topic[0][2:-1])
                else:
                    self.topic_info.append(topic[0][1:-1])
                self.topic_local_id.append(local_id)
                start_topic = end_topic
                end_topic = topic.start() + char_walker - pat_len_before_ann

                if start_topic != end_topic or len(self.topic_info) >= 2:
                    if (
                        len(self.topic_info) >= 2
                    ):  # as we are ignoring the self.topic[0]
                        if start_topic < end_topic:
                            span = CrossVolSpan(
                                self.vol_walker + 1, start_topic, end_topic
                            )
                            self.current_topic_id.append(
                                (
                                    self.topic_local_id[-2],
                                    {"work_id": self.topic_info[-2], "span": span},
                                )
                            )  # -2 as we need the secondlast item
                    else:
                        self.current_topic_id.append(
                            (
                                self.topic_local_id[-1],
                                {
                                    "work_id": self.topic_info[-1],
                                    "span": CrossVolSpan(
                                        self.vol_walker + 1, start_topic, end_topic
                                    ),
                                },
                            )
                        )
                    self.topic_id.append(self.current_topic_id)
                    self.current_topic_id = []
                    if self.sub_topic_Id and end_sub_topic < end_topic:
                        self.sub_topic_Id.append(
                            (
                                self.sub_topic_local_id[-1],
                                {
                                    "work_id": self.sub_topic_info[-1],
                                    "span": CrossVolSpan(
                                        self.vol_walker + 1, end_sub_topic, end_topic
                                    ),
                                },
                            )
                        )
                    self.sub_topic.append(self.sub_topic_Id)
                    self.sub_topic_Id = []
                    if self.sub_topic_Id:
                        last = self.sub_topic_info[-1]
                        self.sub_topic_info = []
                        self.sub_topic_info.append(last)

            if re.search(
                pat_list["error_pattern"], line
            ):  # checking current line contain error annotation or not
                errors = re.finditer(pat_list["error_pattern"], line)
                for error in errors:
                    local_id, span, suggestion = self.parse_payload_ann(
                        error, pat_list, char_walker, line
                    )
                    cur_vol_error_id.append(
                        (local_id, Correction(span, correction=suggestion))
                    )

            if re.search(
                pat_list["archaic_word_pattern"], line
            ):  # checking current line contain error annotation or not
                archaics = re.finditer(pat_list["archaic_word_pattern"], line)
                for archaic in archaics:
                    local_id, span, modern_word = self.parse_payload_ann(
                        archaic, pat_list, char_walker, line
                    )
                    cur_vol_archaic_id.append(
                        (local_id, Archaic(span, modern=modern_word))
                    )

            if re.search(pat_list["abs_er_pattern"], line):
                abs_ers = re.finditer(pat_list["abs_er_pattern"], line)
                for abs_er in abs_ers:
                    pat_len_before_ann = self.search_before(abs_er, pat_list, line)
                    local_id = self.get_local_id(abs_er)
                    start_abs_er = abs_er.start() + char_walker - pat_len_before_ann
                    if local_id:
                        end_abs_er = (
                            start_abs_er + len(abs_er[0][2:-1]) - 1
                        )  # 3 is minus as two border bracket and local id
                    else:
                        end_abs_er = start_abs_er + len(abs_er[0][1:-1]) - 1
                    cur_vol_abs_er_id.append(
                        (local_id, ErrorCandidate(Span(start_abs_er, end_abs_er)))
                    )

            if re.search(pat_list["note_pattern"], line):
                notes_in_line = re.finditer(pat_list["note_pattern"], line)
                for notes in notes_in_line:
                    pat_len_before_ann = self.search_before(notes, pat_list, line)
                    local_id = self.get_local_id(notes)
                    note = notes.start() + char_walker - pat_len_before_ann
                    note_id.append((local_id, {_attr_names.SPAN: Span(note, note)}))

            if re.search(pat_list["start_cit_pattern"], line):
                start_cits = re.finditer(pat_list["start_cit_pattern"], line)
                for start_cit in start_cits:
                    start_cit_patterns.append(
                        self.parse_start_ann(start_cit, pat_list, char_walker, line)
                    )

            if re.search(pat_list["end_cit_pattern"], line):
                end_cits = re.finditer(pat_list["end_cit_pattern"], line)
                for end_cit in end_cits:
                    end_cit_patterns.append(
                        self.parse_end_ann(end_cit, pat_list, char_walker, line)
                    )

            if re.search(pat_list["start_sabche_pattern"], line):
                start_sabches = re.finditer(pat_list["start_sabche_pattern"], line)
                for start_sabche in start_sabches:
                    start_sabche_pattern.append(
                        self.parse_start_ann(start_sabche, pat_list, char_walker, line)
                    )

            if re.search(pat_list["end_sabche_pattern"], line):
                end_sabches = re.finditer(pat_list["end_sabche_pattern"], line)
                for end_sabche in end_sabches:
                    end_sabche_pattern.append(
                        self.parse_end_ann(end_sabche, pat_list, char_walker, line)
                    )

            if re.search(pat_list["start_tsawa_pattern"], line):
                start_tsawas = re.finditer(pat_list["start_tsawa_pattern"], line)
                for start_tsawa in start_tsawas:
                    start_tsawa_pattern.append(
                        self.parse_start_ann(start_tsawa, pat_list, char_walker, line)
                    )

            if re.search(pat_list["end_tsawa_pattern"], line):
                end_tsawas = re.finditer(pat_list["end_tsawa_pattern"], line)
                for end_tsawa in end_tsawas:
                    end_tsawa_pattern.append(
                        self.parse_end_ann(end_tsawa, pat_list, char_walker, line)
                    )

            if re.search(pat_list["start_yigchung_pattern"], line):
                start_yigchungs = re.finditer(pat_list["start_yigchung_pattern"], line)
                for start_yigchung in start_yigchungs:
                    start_yigchung_pattern.append(
                        self.parse_start_ann(
                            start_yigchung, pat_list, char_walker, line
                        )
                    )

            if re.search(pat_list["end_yigchung_pattern"], line):
                end_yigchungs = re.finditer(pat_list["end_yigchung_pattern"], line)
                for end_yigchung in end_yigchungs:
                    end_yigchung_pattern.append(
                        self.parse_end_ann(end_yigchung, pat_list, char_walker, line)
                    )

            if re.search(pat_list["start_durchen_pattern"], line):
                start_durchens = re.finditer(pat_list["start_durchen_pattern"], line)
                for start_durchen in start_durchens:
                    start_durchen_pattern.append(
                        self.parse_start_ann(start_durchen, pat_list, char_walker, line)
                    )

            if re.search(pat_list["end_durchen_pattern"], line):
                end_durchens = re.finditer(pat_list["end_durchen_pattern"], line)
                for end_durchen in end_durchens:
                    end_durchen_pattern.append(
                        self.parse_end_ann(end_durchen, pat_list, char_walker, line)
                    )

            pat_len_before_ann = self.total_pattern(pat_list, line)
            if line:
                end_line = start_line + length - pat_len_before_ann - 1
            else:
                end_line = start_line + length - 1
            char_walker = end_line + 2
            base_line = self.base_extract(pat_list, line) + "\n"
            self.base_text += base_line

            if idx == n_line - 1:  # Last line case
                start_page = end_page
                start_topic = end_topic
                start_sub_topic = end_sub_topic
                if self.sub_topic_Id:
                    self.sub_topic_Id.append(
                        (
                            self.sub_topic_local_id[-1]
                            if self.sub_topic_local_id
                            else None,
                            {
                                "work_id": self.sub_topic_info[-1]
                                if self.sub_topic_info
                                else None,
                                "span": CrossVolSpan(
                                    self.vol_walker + 1,
                                    start_sub_topic,
                                    char_walker - 2,
                                ),
                            },
                        )
                    )
                if self.topic_info:
                    self.current_topic_id.append(
                        (
                            self.topic_local_id[-1],
                            {
                                "work_id": self.topic_info[-1],
                                "span": CrossVolSpan(
                                    self.vol_walker + 1, start_topic, char_walker - 2
                                ),
                            },
                        )
                    )
                if pg_tid:
                    cur_vol_pages.append(
                        (
                            pg_tid[-1],
                            Page(
                                Span(start_page, char_walker - 2),
                                imgnum=pg_ann[-1],
                                page_ref=pg_refs[-1],
                            ),
                        )
                    )
                self.page.append(cur_vol_pages)
                self.error_id.append(cur_vol_error_id)
                cur_vol_error_id = []
                self.abs_er_id.append(cur_vol_abs_er_id)
                self.archaic_word_id.append(cur_vol_archaic_id)
                cur_vol_archaic_id = []
                cur_vol_abs_er_id = []
                self.notes_id.append(note_id)
                note_id = []
                self.author.append(author_titles)
                self.book_title.append(book_titles)
                self.book_number.append(book_numbers)
                self.poti_title.append(poti_titles)
                self.chapter_title.append(chapter_titles)
                self.vol_walker += 1

        if num_vol == self.vol_walker:  # checks the last volume
            self.topic_id.append(self.current_topic_id)
            self.current_topic_id = []
            self.sub_topic.append(self.sub_topic_Id)

        self.citation_pattern.append(
            self.merge(start_cit_patterns, end_cit_patterns)
        )  # The starting and ending of  citation pattern is merged
        self.sabche_pattern.append(
            self.merge(start_sabche_pattern, end_sabche_pattern)
        )  # The starting and ending of  sabche pattern is merged
        self.tsawa_pattern.append(
            self.merge(start_tsawa_pattern, end_tsawa_pattern)
        )  # The starting and ending of  tsawa pattern is merged
        self.yigchung_pattern.append(
            self.merge(start_yigchung_pattern, end_yigchung_pattern)
        )  # The starting and ending of  yigchung pattern is merged
        self.durchen_pattern.append(
            self.merge(start_durchen_pattern, end_durchen_pattern)
        )  # The starting and ending of  yigchung pattern is merged

    def get_result(self):

        if len(self.topic_id) > 1:
            if self.topic_id[0][0][1]["work_id"] == self.topic_id[1][0][1]["work_id"]:
                self.topic_id = self.topic_id[1:]
                self.sub_topic = self.sub_topic[1:]
        self.sub_topic = self.__final_sub_topic(self.sub_topic)
        result = {
            AnnType.book_title: self.book_title,
            AnnType.book_number: self.book_number,
            AnnType.author: self.author,
            AnnType.poti_title: self.poti_title,
            AnnType.chapter: self.chapter_title,
            AnnType.citation: self.citation_pattern,
            AnnType.pagination: self.page,  # page variable format (start_index,end_index,pg_Info,pg_ann)
            AnnType.topic: self.topic_id,
            AnnType.sub_topic: self.sub_topic,
            AnnType.sabche: self.sabche_pattern,
            AnnType.tsawa: self.tsawa_pattern,
            AnnType.yigchung: self.yigchung_pattern,
            AnnType.correction: self.error_id,
            AnnType.error_candidate: self.abs_er_id,
            AnnType.peydurma: self.notes_id,
            AnnType.archaic: self.archaic_word_id,
            AnnType.durchen: self.durchen_pattern,
        }

        return result

    def __final_sub_topic(self, sub_topics):
        """ It include all the sub topic belonging in one topic in a list.

        Args:
            sub_topic (list): It contains all the sub topic annotation's starting and ending index along with sub-topic info.

        Return:
            result (list): It contains list in which sub topic belonging to same topics are grouped together.
        """
        result = []
        cur_topic = []
        cur_sub = []
        sub_topic = sub_topics
        walker = 0
        for i in range(0, len(sub_topic)):
            if len(sub_topic[i]) != 0:
                cur_sub.append(sub_topic[i][0])
                for walker in range(1, len(sub_topic[i])):
                    if (
                        sub_topic[i][walker][1]["work_id"]
                        == sub_topic[i][walker - 1][1]["work_id"]
                    ):
                        cur_sub.append(sub_topic[i][walker])
                    else:
                        cur_topic.append(cur_sub)
                        cur_sub = []
                        cur_sub.append(sub_topic[i][walker])
                if cur_sub:
                    cur_topic.append(cur_sub)
                    cur_sub = []
            else:
                cur_topic.append(cur_sub)
                cur_sub = []
            result.append(cur_topic)
            cur_topic = []
        return result

    def get_base_text(self):
        base_text = self.base_text.strip()
        self.base_text = ""

        return base_text

    def create_opf(self, input_path, id_=None, **kwargs):
        input_path = Path(input_path)
        self._build_dirs(input_path, id_=id_)
        self.metadata = self._load_metadata()
        vol2fn_manager = Vol2FnManager(self.metadata)

        for m_text, vol_name, n_vol in self.get_input(input_path):
            vol_id = vol2fn_manager.get_vol_id(vol_name)
            print(f"[INFO] parsing Vol {vol_name} ...")
            self.build_layers(m_text, n_vol)
            if "is_text" in kwargs:
                if kwargs["is_text"]:
                    continue

            base_text = self.get_base_text()
            (self.dirs["opf_path"] / "base" / f"{vol_id}.txt").write_text(base_text)

        # save pecha layers
        layers = self.get_result()
        for vol_layers, vol_id in self.format_layer(layers):
            if vol_id:
                print(f"[INFO] Creating layers for {vol_id} ...")
                vol_layer_path = self.dirs["layers_path"] / vol_id
                vol_layer_path.mkdir(exist_ok=True)
            else:
                print("[INFO] Creating index layer for Pecha ...")

            for layer, ann in vol_layers.items():
                if layer == "index":
                    layer_fn = self.dirs["opf_path"] / f"{layer}.yml"
                else:
                    layer_fn = vol_layer_path / f"{layer}.yml"
                dump_yaml(ann, layer_fn)

        self._save_metadata(vol2fn=dict(vol2fn_manager.vol2fn))


class HFMLTextFromatter(HFMLFormatter):
    def __init_(self, output_path="./output", is_book=False, text_id=None):
        super().__init__(output_path=output_path, is_book=is_book)
        self.text_id = text_id

    def from_yaml(self, yml_path):
        return yaml.safe_load(yml_path.read_text(encoding="utf-8"))

    def get_offset(self, index_layer, text_id):
        offset = {}
        for text_uuid, text in index_layer["annotations"].items():
            if text["work_id"] == text_id:
                for span in text["span"]:
                    offset[span["vol"]] = span["start"]
                return offset
        return offset

    def get_text_vol_ids(self, text_opf_path):
        base_layers = list((text_opf_path / "base").iterdir())
        base_layers.sort()
        text_vol_ids = [text_id.stem for text_id in base_layers]
        return text_vol_ids

    def get_old_text_base(pecha_idx, pecha_base_text, text_id, text_vol_id):
        vol_num = int(text_vol_id[1:])
        for text_uuid, text in pecha_idx["annotations"].items():
            if text["word_id"] == text_id:
                for vol_span in text["span"]:
                    if vol_span["vol"] == vol_num:
                        return pecha_base_text[vol_span["start"] : vol_span["end"]]
        return pecha_base_text

    def update_base_text(self, pecha_opf_path, text_opf_path, pecha_idx, text_id):
        text_vol_ids = self.get_text_vol_ids(text_opf_path)
        for text_vol_id in text_vol_ids:
            pecha_base_text = (pecha_opf_path / f"base/{text_vol_id}.txt").read_text(
                encoding="utf-8"
            )
            old_text_base = self.get_old_base_text(
                pecha_idx, pecha_base_text, text_id, text_vol_id
            )
            new_text_base = (text_opf_path / f"base/{text_vol_id}.txt").read_text(
                encoding="utf-8"
            )
            new_base_text = pecha_base_text.replace(old_text_base, new_base_text)
            yield text_vol_id, new_base_text
            (pecha_opf_path / f"base/{text_vol_id}.txt").write_text(
                new_base_text, encoding="utf-8"
            )
        print("INFO: base text updated")

    def get_new_ann(self, cur_pecha_ann, cur_text_ann, cur_vol_offset):
        cur_pecha_ann["span"]["start"] = cur_text_ann["span"]["start"] + cur_vol_offset
        cur_pecha_ann["span"]["end"] = cur_text_ann["span"]["end"] + cur_vol_offset
        return cur_pecha_ann

    def get_new_layer(self, cur_vol_pecha_layer, cur_vol_text_layer, cur_vol_offset):
        for ann_id, local_id in cur_vol_text_layer["local_ids"]:
            pecha_ann_id = cur_vol_pecha_layer["local_ids"].key(local_id)
            cur_text_ann = cur_vol_text_layer["annotations"][ann_id]
            cur_pecha_ann = cur_vol_pecha_layer["annotations"][pecha_ann_id]
            cur_vol_pecha_layer["annotations"][pecha_ann_id] = self.get_new_ann(
                cur_pecha_ann, cur_text_ann, cur_vol_offset
            )
        return cur_vol_pecha_layer

    def update_layers(self, pecha_opf_path, text_opf_path, vol_wise_offset):
        text_vol_ids = self.get_text_vol_ids(text_opf_path)
        for text_vol_id in text_vol_ids:
            layers_to_update = list((text_opf_path / f"layers/{text_vol_id}").iterdir())
            for layer_to_update in layers_to_update:
                cur_vol_pecha_layer = self.from_yaml(
                    (
                        pecha_opf_path
                        / f"layers/{text_vol_id}/{layer_to_update.stem}.yml"
                    )
                )
                cur_vol_text_layer = self.from_yaml(layer_to_update)
                cur_vol_offset = vol_wise_offset[text_vol_id]
                new_layer = self.get_new_layer(
                    cur_vol_pecha_layer, cur_vol_text_layer, cur_vol_offset
                )
                new_layer = yaml.safe_load(new_layer)
                (
                    pecha_opf_path / f"layers/{text_vol_id}/{layer_to_update.stem}.yml"
                ).write_text(new_layer, encoding="utf-8")
        print("INFO: layers updated")

    def save_text_to_pecha(self, pecha_opf_path, text_opf_path, text_id):
        pecha_idx = self.from_yaml((pecha_opf_path / "index.yml"))
        for vol_id, new_base_text in self.update_base_text(
            pecha_opf_path, text_opf_path, pecha_idx, text_id
        ):
            (pecha_opf_path / f"base/{vol_id}.txt").write_text(
                new_base_text, encoding="utf-8"
            )
            print(f"INFO: {vol_id} base text updated")
        vol_wise_offset = self.get_offset(pecha_idx, text_id)
        self.update_layers(pecha_opf_path, text_opf_path, vol_wise_offset)
        print(f"{text_id} of {pecha_opf_path.stem} saved...")



import os
from copy import deepcopy
from functools import partial
from pathlib import Path
from re import sub

from bs4 import BeautifulSoup

from .formatter import BaseFormatter
from .layers import *
from .layers import AnnType


class TsadraTemplate:
    """Content tsadra template components.
    """

    author = [
        "credits-page_front-page---text-author",
        "credits-page_front-page---text-author1",
    ]
    book_titles = [
        "credits-page_front-title",
        "tibetan-book-title",
        "credits-page_tibetan-book-title",
    ]
    sub_title = "tibetan-book-sub-title"
    book_number = "credits-page_front-page---book-number"
    commentaries = {
        "verse": [
            "tibetan-commentary-first-line",
            "tibetan-commentary-middle-lines",
            "tibetan-commentary-last-line",
        ],
        "non-verse": [
            "tibetan-commentary-non-indent1",
            "tibetan-commentary-non-indent",
            "tibetan-regular-indented",
            "tibetan-regular-indented1",
        ],
    }
    citation_base = "tibetan-citations-in-verse_tibetan-citations"
    citation_first_line = f"{citation_base}-first-line"
    citation_middle_lines = f"{citation_base}-middle-lines"
    citation_last_line = f"{citation_base}-last-line"
    citation_indent = "tibetan-citations-in-verse_tibetan-regular-indented"
    citation_inline = "tibetan-external-citations"
    tsawa_base = "tibetan-root-text_tibetan-root-text"
    tsawa_first_line = f"{tsawa_base}-first-line"
    tsawa_middle_lines = f"{tsawa_base}-middle-lines"
    tsawa_last_line = f"{tsawa_base}-last-line"
    tsawa_oneliner = [f"{tsawa_base}-1-line-only", ""]
    tsawa_inline = "tibetan-root-text"
    sabches = ["tibetan-sabche", "tibetan-sabche1"]
    yigchung = [
        "tibetan-commentary-small",
        "tibetan-commentary-small-text",
        "tibetan-external-citations-small-letter",
    ]
    footnotes = [
        "tibetan-footnote-reference",
        "tibetan-footnote-reference1",
        "tibetan-footnote",
    ]


class TsadraFormatter(BaseFormatter):
    """
    OpenPecha Formatter for Tsadra DarmaCloud ebooks
    """

    def __init__(self, output_path="./output", metadata=None):
        super().__init__(output_path, metadata=metadata)
        self.base_text = ""
        self.cover_image = ""
        self.walker = 0  # The walker to traverse every character in the pecha
        self.book_title = []  # list variable to store book title index
        self.sub_title = []
        self.book_number = []
        self.poti_title = []
        self.author = []  # list variable to store author annotion index
        self.chapter = []  # list variable to store chapter annotation index
        self.topic = []
        self.sub_topic = []
        self.pagination = []
        self.root_text = []  # list variable to store root text index
        self.citation = []  # list variable to store citation index
        self.sabche = []  # list variable to store sabche index
        self.yigchung = []  # list variable to store yigchung index
        self.footnote_marker = []
        self.footnote_reference = []
        self.footnote = []

    def text_preprocess(self, text):
        return text

    def get_cover_image(self, soup):
        divs = soup.find_all("div", {"epub:type": "cover"})
        if divs:
            return divs[0].img["src"].split("/")[1]
        return ""

    def parse_ann(self, ann, tmp_text, walker, isverse=None):
        if isverse is not None:
            return (
                None,
                ann(Span(walker, len(tmp_text) - 1 + walker), isverse=isverse),
            )
        else:
            return (None, ann(Span(walker, len(tmp_text) - 1 + walker)))

    def get_credit_page(self, p_tag):
        img_tag = p_tag.find("img")
        credit_page_img_name = img_tag["src"].split("/")[1]
        return credit_page_img_name

    def get_footnote_ann(self):
        for marker, ref in zip(self.footnote_marker, self.footnote_reference):
            self.footnote.append((None, Footnote(Span(marker[0], marker[1] - 1), ref)))

    def build_layers(self, html):
        """
        To Build the layer
        """
        soup = BeautifulSoup(html, "html.parser")
        book_title_tmp = ""
        sub_title_tmp = ""
        author_tmp = ""
        chapter_title_tmp = ""
        root_text_tmp = ""
        sabche_tmp = ""
        commentary_tmp = ""
        citation_tmp = ""
        footnote_reference = ""
        footnote_marker = ""
        cover = self.get_cover_image(soup)
        if cover:
            self.cover_image = cover
        ps = soup.find_all("p")
        for p in ps:
            if p["class"][0] in TsadraTemplate.book_titles:
                book_title_tmp = self.text_preprocess(p.text)
                if book_title_tmp:
                    self.book_title.append(
                        self.parse_ann(BookTitle, book_title_tmp, self.walker)
                    )
                    self.base_text += book_title_tmp + "\n"
                    self.walker += len(book_title_tmp) + 1

            elif p["class"][0] == TsadraTemplate.sub_title:
                sub_title_tmp = self.text_preprocess(p.text)
                if sub_title_tmp:
                    self.sub_title.append(
                        self.parse_ann(SubTitle, sub_title_tmp, self.walker)
                    )
                    self.base_text += sub_title_tmp + "\n"
                    self.walker += len(sub_title_tmp) + 1

            elif p["class"][0] == TsadraTemplate.book_number:
                book_num = self.text_preprocess(p.text)
                if book_num:
                    self.book_number.append(
                        self.parse_ann(BookNumber, book_num, self.walker)
                    )
                    self.base_text += book_num + "\n"
                    self.walker += len(book_num) + 1

            elif (
                p["class"][0] in TsadraTemplate.author
            ):  # to get the author annotation index
                author_tmp = self.text_preprocess(p.text)
                self.author.append(self.parse_ann(Author, author_tmp, self.walker))
                self.base_text += author_tmp + "\n"
                self.walker += len(author_tmp) + 1

            elif (
                TsadraTemplate.tsawa_base in p["class"][0]
            ):  # to get the root text or 'tsawa' annotation index (verse form)
                # TODO: one line root text.
                if (
                    p["class"][0] == TsadraTemplate.tsawa_first_line
                    or p["class"][0] == TsadraTemplate.tsawa_middle_lines
                ):
                    root_text_tmp += self.text_preprocess(p.text) + "\n"
                    self.base_text += self.text_preprocess(p.text) + "\n"
                elif p["class"][0] == TsadraTemplate.tsawa_last_line:
                    for s in p.find_all("span"):
                        if "root" in s["class"][0]:
                            root_text_tmp += self.text_preprocess(s.text)
                        else:
                            self.root_text.append(
                                self.parse_ann(
                                    Tsawa, root_text_tmp, self.walker, isverse=True
                                )
                            )
                            self.walker += len(root_text_tmp) + 1
                            root_text_tmp = ""
                            self.walker += len(self.text_preprocess(s.text))
                    if root_text_tmp:
                        self.root_text.append(
                            self.parse_ann(
                                Tsawa, root_text_tmp, self.walker, isverse=True
                            )
                        )
                        self.walker += len(root_text_tmp) + 1
                        root_text_tmp = ""
                    self.base_text += self.text_preprocess(p.text) + "\n"
                else:
                    for s in p.find_all("span"):
                        if "root" in s["class"][0]:
                            root_text_tmp += self.text_preprocess(s.text)
                            self.root_text.append(
                                self.parse_ann(
                                    Tsawa, root_text_tmp, self.walker, isverse=True
                                )
                            )
                            self.walker += len(root_text_tmp) + 1
                            root_text_tmp = ""
                        else:
                            self.walker += len(self.text_preprocess(s.text))
                    self.base_text += self.text_preprocess(p.text) + "\n"

            elif "tibetan-chapter" in p["class"][0]:  # to get chapter title index
                chapter_title_tmp = self.text_preprocess(p.text)
                if chapter_title_tmp:
                    self.chapter.append(
                        self.parse_ann(Chapter, chapter_title_tmp, self.walker)
                    )
                    self.walker += len(chapter_title_tmp) + 1
                    self.base_text += chapter_title_tmp + "\n"

            elif (
                p["class"][0] in TsadraTemplate.commentaries["verse"]
                or p["class"][0] in TsadraTemplate.commentaries["non-verse"]
            ):

                # travesing through commetary which are in verse form
                if p["class"][0] == TsadraTemplate.commentaries["verse"]:
                    commentary_tmp += self.text_preprocess(p.text) + "\n"
                    self.base_text += self.text_preprocess(p.text) + "\n"
                    self.walker += len(commentary_tmp)
                    commentary_tmp = ""
                # travesing through each span of commentary and regular ptag to search annotations
                else:
                    p_tmp = ""
                    p_walker = self.walker
                    for s in p.find_all("span"):
                        try:
                            s["class"][0]
                        except Exception:
                            p_tmp += self.text_preprocess(s.text)
                            continue

                        if (
                            s["class"][0] in TsadraTemplate.yigchung
                        ):  # checking for yigchung annotation
                            if citation_tmp:
                                # citation_tmp += citation_tmp
                                self.citation.append(
                                    self.parse_ann(
                                        Citation, citation_tmp, p_walker, isverse=False
                                    )
                                )
                                p_walker += len(citation_tmp)
                                citation_tmp = ""

                            if sabche_tmp:
                                self.sabche.append(
                                    self.parse_ann(Sabche, sabche_tmp, p_walker)
                                )
                                p_walker += len(sabche_tmp)
                                sabche_tmp = ""

                            self.yigchung.append(
                                self.parse_ann(Yigchung, s.text, p_walker)
                            )
                            p_walker += len(self.text_preprocess(s.text))

                        elif (
                            s["class"][0] == TsadraTemplate.citation_inline
                        ):  # checking for citation annotation
                            citation_tmp += self.text_preprocess(s.text)

                        elif s["class"][0] in TsadraTemplate.footnotes:
                            if len(s.text) < 4 and len(s.text) > 1:
                                footnote_marker = self.text_preprocess(s.text)
                                self.footnote_marker.append(
                                    (p_walker, p_walker + len(footnote_marker))
                                )
                                p_walker += len(footnote_marker)
                            elif len(s.text) > 4:
                                footnote_reference = self.text_preprocess(s.text)
                                if len(footnote_reference) < 100:
                                    self.footnote_reference.append(footnote_reference)
                                p_walker += len(footnote_reference)

                        elif (
                            "sabche" in s["class"][0]
                        ):  # checking for citation annotation
                            sabche_tmp += self.text_preprocess(s.text)

                        elif "front-title" in s["class"][0]:
                            if citation_tmp:
                                self.citation.append(
                                    self.parse_ann(
                                        Citation, citation_tmp, p_walker, isverse=False
                                    )
                                )
                                p_walker += len(citation_tmp)
                                citation_tmp = ""
                            if sabche_tmp:
                                self.sabche.append(
                                    self.parse_ann(Sabche, sabche_tmp, p_walker)
                                )
                                p_walker += len(sabche_tmp)
                                sabche_tmp = ""
                            p_walker += len(self.text_preprocess(s.text))
                        else:
                            if citation_tmp:
                                self.citation.append(
                                    self.parse_ann(
                                        Citation, citation_tmp, p_walker, isverse=False
                                    )
                                )
                                p_walker += len(citation_tmp)
                                citation_tmp = ""

                            if sabche_tmp:
                                self.sabche.append(
                                    self.parse_ann(Sabche, sabche_tmp, p_walker)
                                )
                                p_walker += len(sabche_tmp)
                                sabche_tmp = ""

                            p_walker += len(self.text_preprocess(s.text))

                    # when citation ends the para
                    if citation_tmp:
                        self.citation.append(
                            self.parse_ann(
                                Citation, citation_tmp, p_walker, isverse=False
                            )
                        )
                        p_walker += len(citation_tmp)
                        citation_tmp = ""

                    if sabche_tmp:
                        self.sabche.append(self.parse_ann(Sabche, sabche_tmp, p_walker))
                        p_walker += len(sabche_tmp)
                        sabche_tmp = ""

                    commentary_tmp = self.text_preprocess(p.text) + "\n"
                    self.base_text += self.text_preprocess(commentary_tmp)
                    self.walker += len(commentary_tmp)
                    commentary_tmp = ""
                    p_walker = 0

            elif p["class"][0] in TsadraTemplate.footnotes:
                p_walker = self.walker
                for s in p.find_all("span"):
                    if s["class"][0] in TsadraTemplate.footnotes:
                        if len(s.text) < 4 and len(s.text) > 1:
                            footnote_marker = self.text_preprocess(s.text)
                            self.footnote_marker.append(
                                (p_walker, p_walker + len(footnote_marker))
                            )
                            p_walker += len(footnote_marker)
                        elif len(s.text) > 4:
                            footnote_reference = self.text_preprocess(s.text)
                            if len(footnote_reference) < 100:
                                self.footnote_reference.append(footnote_reference)
                            p_walker += len(footnote_reference)
                    else:
                        p_walker += len(self.text_preprocess(s.text))
                p_tmp = self.text_preprocess(p.text) + "\n"
                self.base_text += self.text_preprocess(p_tmp)
                self.walker += len(p_tmp)

            elif (
                p["class"][0] in TsadraTemplate.sabches
            ):  # checking for sabche annotation
                sabche_tmp = ""
                p_with_sabche_tmp = ""
                sabche_walker = self.walker
                for s in p.find_all("span"):
                    try:
                        s["class"][0]
                    except Exception:
                        p_with_sabche_tmp += self.text_preprocess(s.text)
                        continue

                    if "sabche" in s["class"][0]:
                        sabche_tmp += self.text_preprocess(s.text)

                    else:
                        if sabche_tmp:
                            self.sabche.append(
                                self.parse_ann(Sabche, sabche_tmp, sabche_walker)
                            )
                            sabche_walker = len(sabche_tmp) + sabche_walker
                            sabche_tmp = ""
                        sabche_walker += len(self.text_preprocess(s.text))

                # when sabche ends the para
                if sabche_tmp:
                    self.sabche.append(
                        self.parse_ann(Sabche, sabche_tmp, sabche_walker)
                    )
                    sabche_tmp = ""
                self.walker += len(self.text_preprocess(p.text)) + 1
                self.base_text += self.text_preprocess(p.text) + "\n"
                sabche_walker = 0
            elif (
                TsadraTemplate.citation_base in p["class"][0]
            ):  # checking for citation annotation first two if for verse form and last for non verse
                if (
                    p["class"][0] == TsadraTemplate.citation_first_line
                    or p["class"][0] == TsadraTemplate.citation_middle_lines
                ):
                    citation_tmp += self.text_preprocess(p.text) + "\n"
                    self.base_text += self.text_preprocess(p.text) + "\n"
                elif TsadraTemplate.citation_last_line in p["class"][0]:
                    citation_tmp += self.text_preprocess(p.text) + "\n"
                    self.citation.append(
                        self.parse_ann(
                            Citation, citation_tmp, self.walker, isverse=True
                        )
                    )
                    self.base_text += self.text_preprocess(p.text) + "\n"
                    self.walker += len(citation_tmp)
                    citation_tmp = ""
                elif p["class"][0] == TsadraTemplate.citation_indent:
                    citation_tmp += self.text_preprocess(p.text)
                    self.base_text += self.text_preprocess(p.text) + "\n"
                    self.citation.append(
                        self.parse_ann(
                            Citation, citation_tmp, self.walker, isverse=True
                        )
                    )
                    self.walker += len(citation_tmp) + 1
                    citation_tmp = ""
            else:
                if p.text:
                    self.base_text += self.text_preprocess(p.text) + "\n"
                    self.walker += len(self.text_preprocess(p.text)) + 1

    def get_result(self):
        """
        To return all the result
        """
        self.get_footnote_ann()
        result = {
            AnnType.book_title: [self.book_title],
            AnnType.sub_title: [self.sub_title],
            AnnType.book_number: [self.book_number],
            AnnType.poti_title: [self.poti_title],
            AnnType.author: [self.author],
            AnnType.chapter: [self.chapter],
            AnnType.topic: [self.topic],
            AnnType.sub_topic: [self.sub_topic],
            AnnType.pagination: [self.pagination],
            AnnType.tsawa: [self.root_text],
            AnnType.citation: [self.citation],
            AnnType.sabche: [self.sabche],
            AnnType.yigchung: [self.yigchung],
            AnnType.footnote: [self.footnote],
        }
        return result

    def _get_meta_data(self, ann):
        return ",".join([self.base_text[a[0] : a[1] + 1] for a in ann])

    def get_base_text(self):
        """
        To return base text of each processed page
        """
        return self.base_text

    def get_input(self, input_path):
        def get_prefix(html_paths):
            self.sku = list(html_paths[0].parents)[1].name
            return sorted(html_paths, key=lambda p: len(p.stem))[0].stem

        def semantic_order(sku, html_path):
            html_fn = html_path.stem
            order = html_fn[len(sku) + 1 :]
            if order:
                return int(order)
            else:
                return 0

        html_paths = [
            o
            for o in input_path.iterdir()
            if o.suffix == ".xhtml" and o.stem != "cover"
        ]
        sku_sementic_order = partial(semantic_order, get_prefix(html_paths))
        html_paths = sorted(html_paths, key=sku_sementic_order)
        html_paths.insert(0, input_path / "cover.xhtml")

        for html_fn in html_paths:
            yield Path(html_fn).read_text()

    def create_metadata(self, layers):
        def get_text(span):
            return self.base_text[span["start"] : span["end"] + 1].replace("\n", "")

        meta_data = {}
        meta_data["title"] = get_text(layers["BookTitle"][0][1][0][1]["span"])
        meta_data["authors"] = [
            get_text(span["span"]) for (layer, span) in layers["Author"][0][1]
        ]
        meta_data["sku"] = self.sku
        meta_data["cover"] = self.cover_image
        meta_data["layers"] = [layer for layer in layers if layers[layer]]
        return {"ebook_metadata": meta_data}

    def create_opf(self, input_path, id_):
        input_path = Path(input_path)
        self._build_dirs(input_path, id_=id_)
        (self.dirs["opf_path"] / "base").mkdir(exist_ok=True)

        # cover image path
        image_path = input_path / "image"
        (self.dirs["opf_path"] / "assets").mkdir(exist_ok=True)
        os.system(f"cp -R {image_path} {self.dirs['opf_path']}/assets")

        # parse layers
        for html in self.get_input(input_path):
            self.build_layers(html)

        # save base-text
        (self.dirs["opf_path"] / "base" / "v001.txt").write_text(self.get_base_text())

        # format and save layer
        vol_layer_path = self.dirs["layers_path"] / "v001"
        vol_layer_path.mkdir(exist_ok=True)
        layers = self.get_result()
        for vol_layer, base_id in self.format_layer(layers):
            for layer, ann in vol_layer.items():
                if ann["annotations"]:
                    if layer == "index":
                        layer_fn = self.dirs["opf_path"] / f"{layer}.yml"
                    else:
                        layer_fn = vol_layer_path / f"{layer}.yml"
                    dump_yaml(ann, layer_fn)

        # save metatdata
        meta_data = self.create_metadata(layers)
        meta_fn = self.dirs["opf_path"] / "meta.yml"
        dump_yaml(meta_data, meta_fn)


# if __name__ == "__main__":
#     ebook_path = "./output/demo/src/tsadra_publication/P000112/OEBPS/"
#     opfs_path = "./output/demo/output"
#     formatter = TsadraFormatter(output_path=opfs_path)
#     formatter.create_opf(ebook_path, 112)



import re


class StringChunker:
    def __init__(self, full_string, pattern, target_len = 1500, start_cc = 0, end_cc = 0):
        self.pattern  = pattern
        self.string   = full_string
        self.target_len  = target_len
        self.start_cc = start_cc
        self.end_cc   = end_cc if end_cc > 0 else len(full_string)

    def get_next_chunk_index(self, last_index):
        """
        Returns an index of where the string has the last char, if the char is not found in n char,
        the function doubles n and looks for the first char in that double chunk,
        if the char is still not found it looks for the last space in the original chunk,
        if still no space was found in the chunk than just returns n to cut up the string arbitrarily
        """
        end_cc = min(self.end_cc, last_index + 2 * self.target_len)
        next_target = last_index + self.target_len
        next_idx_before_target_len = -1
        for match in self.pattern.finditer(self.string, last_index, end_cc):
            if match.end() < next_target:
                next_idx_before_target_len = match.end()
            else:
                if next_idx_before_target_len != -1:
                    return next_idx_before_target_len
                return match.end()
        return next_idx_before_target_len if next_idx_before_target_len > 0 else -1

    def get_chunks(self):
        """
        Returns a list of indexes of the string cut up in chunks of a maximum of n length and cut up at the last
        instance of regex
        """
        list_of_index = [self.start_cc]
        string = self.string
        index = self.start_cc
        while self.end_cc - index > self.target_len:
            found = self.get_next_chunk_index(index)
            if found == -1:
                found = self.string.rfind(" ", index+1, index+self.target_len)
            if found == -1:
                found = index+self.target_len
            index = found
            list_of_index.append(index)
        if index != self.end_cc:
            list_of_index.append(self.end_cc)
        return list_of_index


ENG_PATTERN = re.compile(r"\.\s")


class EnglishEasyChunker(StringChunker):
    def __init__(self, full_string, target_len, start_cc = 0, end_cc = 0):
        StringChunker.__init__(
            self,
            full_string,
            ENG_PATTERN,
            target_len,
            start_cc,
            end_cc
        )


TIB_PATTERN = re.compile(r"([སའངགདནབམརལཏ]ོ[་༌]?[།-༔][^ཀ-ཬ]*|(།།|[༎-༒])[^ཀ-ཬ༠-༩]*[།-༔][^ཀ-ཬ༠-༩]*)")


class TibetanEasyChunker(StringChunker):
    def __init__(self, full_string, target_len, start_cc = 0, end_cc = 0):
        StringChunker.__init__(
            self,
            full_string,
            TIB_PATTERN,
            target_len,
            start_cc,
            end_cc
        )



from pathlib import Path


def _mkdir(path):
    if path.is_dir():
        return path
    path.mkdir(exist_ok=True, parents=True)
    return path


# Path
BASE_PATH = _mkdir(Path.home() / ".openpecha")
PECHAS_PATH = _mkdir(BASE_PATH / "pechas")

# Pecha
PECHA_PREFIX = "P"
N_SIG = 6  # no. of significant digits for pecha id

GITHUB_ORG_URL = "https://github.com/OpenPecha-data"

# Work
WORKS_REPO_NAME = "works"
OP2BDRC_WORK_MAPPING_URL = (
    "https://raw.githubusercontent.com/OpenPecha/works/main/mappings/bdrc.csv"
)



import json
import shutil
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Union
from git import Repo

from openpecha import blupdate, config
from openpecha.core import ids
from openpecha.core.annotations import BaseAnnotation, Span
from openpecha.core.layer import Layer, LayerEnum, PechaMetadata, SpanINFO
from openpecha.storages import GithubStorage, Storage
from openpecha.utils import download_pecha, dump_yaml, load_yaml, load_yaml_str


class OpenPecha:
    """
    The main abstract class representing an opf. The methods to implement are:
    - read_meta_file()
    - read_base_file(base_name)
    - read_layers_file(base_name, layer_name)
    - _read_components()
    - read_index_file()
    """

    def __init__(
        self,
        bases: Dict[str, str] = None,
        layers: Dict[str, Dict[LayerEnum, Layer]] = None,
        index: Layer = None,
        meta: PechaMetadata = None,
        metadata: PechaMetadata = None,
        assets: Dict[str, List[Union[str, Path]]] = None,
        components: Dict[str, List[Layer]] = None,
    ):
        self._pecha_id = None
        self.bases = bases if bases else {}
        self.layers = layers if layers else defaultdict(dict)
        self._meta = self.__handle_old_metadata_attr(meta, metadata)
        self._index = index
        self.assets = assets if assets else {}
        self._components = components if components else {}
        self.current_base_order = 1

    def __str__(self):
        return f"OpenPecha:{self.pecha_id}"

    def __handle_old_metadata_attr(self, old, new):
        if not old and not new:
            return

        if new:
            return new

        warnings.warn(
            "use `metadata` attr instead of `meta` to set pecha metadata",
            DeprecationWarning,
            stacklevel=3,
        )
        return old

    @property
    def about(self):
        source_metadata = []
        for val in self.meta.source_metadata.values():
            if not isinstance(val, (str, int)):
                continue
            source_metadata.append(val)
        return ", ".join(source_metadata)

    def reset_base_and_layers(self):
        self.bases = {}
        self.layers = defaultdict(dict)

    @property
    def pecha_id(self) -> str:
        if self._pecha_id:
            return self._pecha_id
        self._pecha_id = self.meta.id
        return self._pecha_id

    @property
    def meta(self) -> PechaMetadata:
        if self._meta:
            return self._meta
        self._meta = PechaMetadata.parse_obj(self.read_meta_file())
        return self._meta

    @property
    def index(self) -> Layer:
        if self._index:
            return self._index
        self._index = Layer.parse_obj(self.read_index_file())
        return self._index

    @property
    def components(self) -> Dict[str, List[LayerEnum]]:
        if self._components:
            return self._components
        self._components = self._read_components()
        return self._components

    def _get_base_name(self) -> str:
        return ids.get_base_id()

    def _set_base_metadata(self, base_name: str, metadata: Dict) -> None:
        metadata.update({"base_file": f"{base_name}.txt"})
        if "bases" not in self.meta:
            self.meta.bases = {}
        self.meta.bases[base_name] = metadata

    def get_base(self, base_name: str) -> str:
        if base_name in self.bases:
            return self.bases[base_name]
        self.bases[base_name] = self.read_base_file(base_name)
        return self.bases[base_name]

    def get_base_metadata(self, base_name: str) -> str:
        self.meta.bases.get(base_name)

    def set_base(self, content: str, base_name: str = None, metadata: Dict = {}) -> str:
        """Create new base with `content` if `base_name` is not
        given otherwise overwrites it and return base_name.
        """
        if base_name and base_name in self.bases:
            blupdate.update_single_base(self, base_name, content)
            self.bases[base_name] = content
        else:
            base_name = self._get_base_name()
            self.bases[base_name] = content
            self._set_base_metadata(base_name, metadata)
        return base_name

    def get_layers(self, base_name: str) -> Layer:
        for layer_name in self.components[base_name]:
            yield self.get_layer(base_name, layer_name)

    def get_layer(self, base_name: str, layer_name: LayerEnum) -> Layer:
        if base_name in self.layers and layer_name in self.layers[base_name]:
            layer = self.layers[base_name][layer_name]
            if layer:
                return layer

        layer_dict = self.read_layers_file(base_name, layer_name.value)
        if layer_dict:
            layer = Layer.parse_obj(layer_dict)
        else:
            layer = Layer(annotation_type=layer_name)
        self.layers[base_name][layer_name] = layer
        return layer

    def update_metadata(self):
        self.meta.update_last_modified_date()

    def set_layer(self, base_name: str, layer: Layer):
        if base_name not in self.bases:
            raise ValueError(f"set base for {base_name} first")

        self.layers[base_name][layer.annotation_type] = layer

    def __find_span_layers(
        self, base_name: str, span: Span, layers: List[LayerEnum]
    ) -> Dict[LayerEnum, List[BaseAnnotation]]:
        result = defaultdict(list)
        for layer_name in layers:
            if layer_name not in self.components[base_name]:
                result[layer_name] = []

            layer = self.get_layer(base_name, layer_name)
            for _, ann in layer.get_annotations():
                is_ann_found = False
                if ann.span.start >= span.start and ann.span.end <= span.end:
                    is_ann_found = True
                elif ann.span.end >= span.start and ann.span.start < span.start:
                    is_ann_found = True
                    ann.span.start = span.start
                elif ann.span.start <= span.end and ann.span.end > span.end:
                    is_ann_found = True
                    ann.span.end = span.end

                if is_ann_found:
                    ann.span.start -= span.start
                    ann.span.end -= span.start
                    result[layer_name].append(ann)

        return result

    def get_span_info(
        self, base_name: str, span: Span, layers: List[LayerEnum] = []
    ) -> SpanINFO:
        base_str = self.get_base(base_name)
        span_str = base_str[span.start : span.end + 1]  # noqa
        layers = layers if layers else self.components[base_name]
        span_layers = self.__find_span_layers(base_name, span, layers)
        return SpanINFO(text=span_str, layers=span_layers, metadata=self.meta)


class OpenPechaFS(OpenPecha):
    """Class to represent opf pecha on file-system.

    Note:
        Either use pecha_id or `path` attribute to create
        instance, `pecha_id` for downloading/updating pecha and
        `path` for local pecha.

    Attributes:
        pecha_id(str): id of openpecha pecha.
        path (str): path to local pecha root or pecha .opf path.
        storage (Storage): storage obj for saving at remote.
    """

    def __init__(
        self, pecha_id: str = None, path: str = None, storage: Storage = None, **kwargs
    ):
        self._opf_path = self.get_opf_path(pecha_id, path)
        self.output_dir = None
        self.storage = storage
        super().__init__(**kwargs)

    @staticmethod
    def get_opf_path(pecha_id, path: str) -> Path:
        """convert pecha path to pecha's opf path"""
        if not pecha_id and not path:
            return

        if not path:
            return download_pecha(pecha_id)

        path = Path(path)
        if path.name.endswith(".opf"):
            return path
        path = path / f"{path.name}.opf"
        if not path.is_dir():
            raise FileNotFoundError(f"Pecha not found at: {path}")
        return path

    @staticmethod
    def _mkdir(path: Path):
        path.mkdir(parents=True, exist_ok=True)
        return path

    @property
    def opf_path(self) -> Path:
        if self._opf_path:
            return self._opf_path
        self._opf_path = self.pecha_path / f"{self.pecha_id}.opf"
        return self._opf_path

    @property
    def pecha_path(self) -> Path:
        return self.output_dir / self.pecha_id

    @property
    def base_path(self) -> Path:
        return self.opf_path / "base"

    @property
    def layers_path(self) -> Path:
        return self.opf_path / "layers"

    @property
    def meta_fn(self) -> Path:
        return self.opf_path / "meta.yml"

    @property
    def index_fn(self) -> Path:
        return self.opf_path / f"{LayerEnum.index.value}.yml"

    @property
    def assets_path(self) -> Path:
        return self.opf_path / "assets"

    def read_base_file(self, base_name: str) -> str:
        return (self.base_path / f"{base_name}.txt").read_text(encoding="utf-8")

    def read_layers_file(
        self, base_name: str, layer_name: LayerEnum
    ) -> Union[Dict, None]:
        layer_fn = self.layers_path / base_name / f"{layer_name}.yml"
        if layer_fn.is_file():
            return load_yaml(layer_fn)

    def read_meta_file(self) -> Dict:
        return load_yaml(self.meta_fn)

    def read_index_file(self) -> Dict:
        if not self.index_fn.is_file():
            raise FileNotFoundError
        return load_yaml(self.index_fn)

    def _read_components(self):
        res = {}
        for vol_dir in self.layers_path.iterdir():
            all_layers = set(layer.value for layer in LayerEnum)
            res[vol_dir.name] = list(
                map(
                    lambda fn: LayerEnum(fn.stem),
                    (
                        layer_fn
                        for layer_fn in vol_dir.iterdir()
                        if layer_fn.stem in all_layers
                    ),
                )
            )
        return res

    def save_meta(self):
        dump_yaml(json.loads(self.meta.json()), self.meta_fn)

    def save_single_base(self, base_name: str, content: str = None):
        if not content:
            content = self.bases[base_name]
        base_fn = self._mkdir(self.base_path) / f"{base_name}.txt"
        base_fn.write_text(content)

    def save_base(self):
        for base_name, content in self.bases.items():
            self.save_single_base(base_name, content)

    def save_layer(self, base_name: str, layer_name: LayerEnum, layer: Layer):
        layer_fn = self._mkdir(self.layers_path / base_name) / f"{layer_name.value}.yml"
        dump_yaml(json.loads(layer.json()), layer_fn)
        return layer_fn

    def save_layers(self):
        for base_name, base_layers in self.layers.items():
            for layer_name, layer in base_layers.items():
                self.save_layer(base_name, layer_name, layer)

    def save_index(self):
        try:
            dump_yaml(json.loads(self.index.json()), self.index_fn)
        except FileNotFoundError:
            pass

    def save_assets(self):
        for assets_type, content in self.assets.items():
            assets_type_dir = self.assets_path / assets_type
            assets_type_dir.mkdir(parents=True, exist_ok=True)
            for asset_fn in content:
                asset_fn = Path(asset_fn)
                dest_fn = assets_type_dir / asset_fn.name
                shutil.copyfile(str(asset_fn), str(dest_fn))

    def save(self, output_path: Union[str, Path] = config.PECHAS_PATH):
        self.output_dir = Path(output_path)
        self._opf_path = None

        self.save_base()
        self.save_layers()
        self.save_index()
        self.save_meta()
        self.save_assets()
        return self.opf_path

    def update_base(self, base_name: str, content: str):
        self.set_base(content, base_name)
        self.save_single_base(base_name)
        self.update_metadata()

    def update_layer(self, base_name: str, layer_name: LayerEnum, layer: Layer):
        old_layer = self.get_layer(base_name, layer_name)
        old_layer.bump_revision()
        old_layer.annotations = layer.annotations
        self.save_layer(base_name, layer_name, old_layer)
        self.update_metadata()

    def reset_layer(self, base_name: str, layer_name: LayerEnum):
        layer = self.get_layer(base_name, layer_name)
        layer.reset()
        self.save_layer(base_name, layer_name, layer)
        self.layers[base_name][layer_name] = None

    def reset_layers(self, base_name: str, exclude: List[LayerEnum] = []):
        for layer_name in self.components[base_name]:
            if layer_name in exclude:
                continue
            self.reset_layer(base_name, layer_name)

    def publish(self):
        if not self.storage:
            self.storage = GithubStorage()
        self.storage.add_dir(path=self.pecha_path, description=self.about)

    def remove(self):
        self.storage.remove_dir_with_path(name=self.pecha_path)


class OpenPechaBareGitRepo(OpenPecha):
    """Class to represent opf pecha in a bare git repo on the file-system.
    """

    def __init__(
        self, pecha_id: str = None, path: str = None, revision: str = "HEAD", repo=None, **kwargs
    ):
        super().__init__(**kwargs)
        self._pecha_id = pecha_id
        self.rev = revision
        if repo is not None:
            self.repo = repo
        else:
            self.repo = Repo(path)

    def read_file_content(self, oppath):
        return self.repo.git.show(f"{self.rev}:{self._pecha_id}.opf/" + oppath)

    def read_file_content_yml(self, oppath):
        ymlstr = self.repo.git.show(f"{self.rev}:{self._pecha_id}.opf/" + oppath)
        return load_yaml_str(ymlstr)

    def _list_paths(self):
        """
        Getting all the files in the bare repo
        """
        files = self.repo.git.ls_tree(r=self.rev).split("\n")
        # removing the xxx.opf at the beginning
        files = [file.split("\t")[-1] for file in files]
        files = [
            file[len(self._pecha_id) + 5 :]
            for file in files
            if file.startswith(self.lname + ".")
            and (file.endswith(".txt") or file.endswith(".yml"))
        ]
        return files

    def _read_components(self):
        """
        get all bases and layers
        """
        paths = self._list_paths()
        res = {}

        for f in sorted(paths):
            path = f.split("/")
            if len(path) > 1:
                basename = path[-2]
                layername = pathlib.Path(path[-1]).stem
                if basename not in res:
                    res[basename] = []
                res[basename].append(layername)
        return res

    def read_base_file(self, base_name: str) -> str:
        return self.read_file_content("base/" + base_name + ".txt")

    def read_layers_file(
        self, base_name: str, layer_name: LayerEnum
    ) -> Union[Dict, None]:
        return self.read_file_content_yml("layers/" + base_name + "/" + layer_name + ".yml")

    def read_meta_file(self) -> Dict:
        return self.read_file_content_yml("meta.yml")

    def read_index_file(self) -> Dict:
        return self.read_file_content_yml("index.yml")




import os
import re
import shutil
import zipfile
from enum import Enum
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from openpecha.formatters.layers import AnnType
from openpecha.utils import load_yaml

from .serialize import Serialize


class TsadraTemplateCSSClasses(Enum):
    cover_title = 'credits-page_front-title'
    book_title = 'tibetan-book-title'
    sub_title = 'tibetan-book-sub-title'
    book_number = 'tibetan-book-number'
    chapter = 'tibetan-chapters'
    tsawa_inline = 'tibetan-root-text'
    tsawa_verse = 'tibetan-root-text-in-verse'
    citation_inline = 'tibetan-citations'
    citation_verse = 'tibetan-citations-in-verse'
    citation_prose = 'tibetan-citations-prose'
    sabche = 'tibetan-sabche1'
    yigchung = 'tibetan-commentary-small'

class Tsadra_template:
    """Variables are important components of tsadra epub template."""

    # SP = Start_Payload
    span_EP = "</span>"
    para_EP = "</p>"
    ft = '<span class="front-title">'
    cover_title_SP = f'<span class="{TsadraTemplateCSSClasses.cover_title.value}">'
    book_title_SP = f'<span class="{TsadraTemplateCSSClasses.book_title.value}">'
    sub_title_SP = f'<span class="{TsadraTemplateCSSClasses.sub_title.value}">'
    book_number_SP = f'<p class="{TsadraTemplateCSSClasses.book_number.value}">{ft}'
    credit_page_SP = (
        '<p class="credits-page_epub-edition-line"><span class="credits-regular">'
    )
    author_SP = '<p class="text-author"><span class="front-page---text-titles">'
    chapter_SP = f'<span class="{TsadraTemplateCSSClasses.chapter.value}">'
    tsawa_inline_SP = f'<span class="{TsadraTemplateCSSClasses.tsawa_inline.value}">'
    tsawa_verse_SP = f'<span class="{TsadraTemplateCSSClasses.tsawa_verse.value}">'
    citation_inline_SP = f'<span class="{TsadraTemplateCSSClasses.citation_inline.value}">'
    citation_verse_SP = f'<span class="{TsadraTemplateCSSClasses.citation_verse.value}">'
    citation_prose_sp = f'<span class="{TsadraTemplateCSSClasses.citation_prose.value}">'
    sabche_SP = f'<span class="{TsadraTemplateCSSClasses.sabche.value}">'
    yigchung_SP = f'<span class="{TsadraTemplateCSSClasses.yigchung.value}">'
    footnote_marker_SP = '<span class="tibetan-footnote-marker"'
    footnote_EP = "</span></a>"
    footnote_reference_SP = '<span class="tibetan-footnote-reference"'

    toc_xpaths = {
        "book-number": f"//*[@class='{TsadraTemplateCSSClasses.book_number.value}']",
        "chapter": f"//*[@class='{TsadraTemplateCSSClasses.chapter.value}']",
        "sabche": f"//*[@class='{TsadraTemplateCSSClasses.sabche.value}' or @class='tibetan-sabche']",
    }
    toc_levels = {"1": "book-number", "2": "chapter", "3": "sabche"}
    book_title_Xpath = f"//*[@class='{TsadraTemplateCSSClasses.book_title.value}']"


class EpubSerializer(Serialize):
    """Epub serializer class for OpenPecha."""

    def __get_adapted_span(self, span, base_id):
        """Adapts the annotation span to base-text of the text

        Adapts the annotation span, which is based on volume base-text
        to text base-text.

        Args:
            span (dict): span of a annotation, eg: {start:, end:}
            base_id (str): id of vol, where part of the text exists.

        Returns:
            adapted_start (int): adapted start based on text base-text
            adapted_end (int): adapted end based on text base-text

        """
        adapted_start = span["start"] - self.text_spans[base_id]["start"]
        adapted_end = span["end"] - self.text_spans[base_id]["start"]
        return adapted_start, adapted_end

    def get_citation_sp(self, css_class_name):
        """Return start payload for citation according to class name

        Args:
            css_class_name (str): css class name to differentiate different types of citations

        Returns:
            str: citation start payload
        """
        start_payload = Tsadra_template.citation_inline_SP
        if css_class_name:
            css_class_enum = TsadraTemplateCSSClasses(css_class_name)
            if css_class_enum == TsadraTemplateCSSClasses.citation_verse:
                start_payload = Tsadra_template.citation_verse_SP
            elif css_class_enum == TsadraTemplateCSSClasses.citation_prose:
                start_payload = Tsadra_template.citation_prose_sp
        return start_payload


    def get_tsawa_sp(self, css_class_name):
        """Return start payload for tsawa according to class name

        Args:
            css_class_name (str): css class name to differentiate different types of tsawa

        Returns:
            str: tsawa start payload
        """
        start_payload = Tsadra_template.tsawa_inline_SP
        if css_class_name:
            css_class_enum = TsadraTemplateCSSClasses(css_class_name)
            if css_class_enum == TsadraTemplateCSSClasses.tsawa_verse:
                start_payload = Tsadra_template.tsawa_verse_SP
        return start_payload

    def apply_annotation(self, base_id, ann, uuid2localid):
        """Applies annotation to specific volume base-text, where part of the text exists.

        Args:
            base_id (str): id of vol, where part of the text exists.
            ann (dict): annotation of any type.

        Returns:
            None

        """
        only_start_ann = False
        start_payload = "("
        end_payload = ")"
        if ann["type"] == AnnType.correction:
            start_payload = "("
            end_payload = f',{ann["correction"]})'
        elif ann["type"] == AnnType.peydurma:
            start_payload = "#"
            only_start_ann = True
        elif ann["type"] == AnnType.error_candidate:
            start_payload = "["
            end_payload = "]"
        elif ann["type"] == AnnType.book_title:
            start_payload = Tsadra_template.book_title_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.sub_title:
            start_payload = Tsadra_template.sub_title_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.book_number:
            start_payload = Tsadra_template.book_number_SP
            end_payload = f"{Tsadra_template.span_EP}{Tsadra_template.para_EP}"
        elif ann["type"] == AnnType.author:
            start_payload = Tsadra_template.author_SP
            end_payload = f"{Tsadra_template.span_EP}{Tsadra_template.para_EP}"
        elif ann["type"] == AnnType.chapter:
            start_payload = Tsadra_template.chapter_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.tsawa:
            css_class_name = self.get_css_class_name(ann)
            start_payload = self.get_tsawa_sp(css_class_name)
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.citation:
            css_class_name = self.get_css_class_name(ann)
            start_payload = self.get_citation_sp(css_class_name)
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.sabche:
            start_payload = Tsadra_template.sabche_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.yigchung:
            start_payload = Tsadra_template.yigchung_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.footnote:
            start_payload = f'<a href="#fr{ann["id"]}">{Tsadra_template.footnote_marker_SP} id="fm{ann["id"]}">'
            end_payload = Tsadra_template.footnote_EP

        start_cc, end_cc = self.__get_adapted_span(ann["span"], base_id)
        self.add_chars(base_id, start_cc, True, start_payload)
        if not only_start_ann:
            self.add_chars(base_id, end_cc, False, end_payload)

    def p_tag_adder(self, body_text):
        """Add p tag to lines where it is missing

        Args:
            body_text (str): body tag of ebook after layers are applied

        Returns:
            str: body tag with proper p tags
        """
        new_body_text = ""
        body_text = re.sub(r"\n</span>", "\n</span>\n", body_text)
        paras = body_text.split("\n")
        para_flag = False
        cur_span_payload = ""
        cur_para = ""
        for para in paras:
            if "<p" not in para:
                if re.search("<span.+?</span>", para):
                    new_body_text += f"<p>{para}</p>"
                elif "<span" in para and not para_flag:
                    cur_span_payload = re.search("<span .+>", para)[0]
                    cur_para += f"<p>{para}</span></p>"
                    para_flag = True
                elif para_flag and "</span>" not in para:
                    cur_para += (
                        f'<p>{cur_span_payload[:-2]}-middle-line">{para}</span></p>'
                    )
                elif "</span>" in para:
                    cur_para += f'<p>{cur_span_payload[:-2]}-last-line">{para}</p>'
                    new_body_text += cur_para
                    cur_para = ""
                    para_flag = False
                    cur_span_payload = ""
                else:
                    new_body_text += f"<p>{para}</p>"
            else:
                new_body_text += para
        return new_body_text

    def is_title(self, p_tag):
        """Check if a p tag is title type annotation such as title, subtitle, chapter or book-number

        Args:
            p_tag (str): p tag of a body text

        Returns:
            boolean: True if p tag is title type annotation else False
        """
        if "title" in p_tag or "chapter" in p_tag or "number" in p_tag:
            return True
        else:
            return False

    def is_sabche_only(self, p_tag):
        """Check if p tag is non inline sabche or not

        Args:
            p_tag (str): p tag of a body text

        Returns:
            boolean: True if p tag is non inline sache else False
        """
        if re.search('<p><span class="tibetan-sabche1">', p_tag):
            return True
        else:
            return False

    def rm_indentation(self, p_tag):
        """Return p tag with non indented style

        Args:
            p_tag (str): p tag of a body text

        Returns:
            str: p tag with non indented style
        """
        if self.is_sabche_only(p_tag):
            p_tag = p_tag.replace(
                '<p><span class="tibetan-sabche1">', '<p><span class="tibetan-sabche">'
            )
        if len(self.get_p_text(p_tag)) > 50:
            p_tag = re.sub("<p>", '<p class="tibetan-commentary-non-indent">', p_tag)
        else:
            p_tag = p_tag.replace("<p>", '<p class="tibetan-regular-indented">')
        return p_tag

    def add_indentation(self, p_tag):
        """Add indentation style to p tag

        Args:
            p_tag (str): p tag of boody text

        Returns:
            str: p tag with indentation style
        """
        p_tag = re.sub("<p>", '<p class="tibetan-regular-indented">', p_tag)
        return p_tag

    def is_annotated_p_tag(self, p_tag):
        """Check a p tag is annotated p tag or not

        Args:
            p_tag (str): p tag of a body text

        Returns:
            boolean: True if p tag is annotated else False
        """
        if re.search("<p><span.*?>.+?</span></p>", p_tag) or re.search(
            '<p class=".+"><span class=".+">.+?</span></p>', p_tag
        ):
            return True
        else:
            return False

    def get_p_text(self, p_tag):
        """Return text included in p tag

        Args:
            p_tag (str): p tag of body text

        Returns:
            str: text of p tag
        """
        para_tag = re.search(r"<p.*?>(<span.*?>)?(.+?)(</span>)?</p>", p_tag)
        para_text = para_tag.group(2)
        return para_text

    def get_p_tags(self, body_text):
        """Return all the p tags in body text

        Args:
            body_text (str): body text of the ebook

        Returns:
            list: p tags in body text
        """
        p_tags = re.split("(<p.*?>.+?</p>)", body_text)
        return p_tags[1::2]

    def has_yigchung(self, p_tag):
        """Check if p tag contains yigchung or not

        Args:
            p_tag (str): p tag of a body

        Returns:
            boolean: True of yigchung exist in p tag else False
        """
        if Tsadra_template.yigchung_SP in p_tag:
            return True
        else:
            return False

    def get_front_page(self):
        """Construct the front page of the book including title, subtitle, authors and credit page

        Returns:
            str: front page content
        """
        front_page = ""
        source_metadata = self.meta.get("source_metadata", {})
        credit_pg_name = source_metadata.get("credit", "")
        credit_pg_path = self.opf_path / f"assets/image/{credit_pg_name}"
        book_title_tag = f'<p class="tibetan-regular-indented">{Tsadra_template.cover_title_SP}{source_metadata.get("title", "")}</span></p>\n'
        sub_title_tag = f'<p class="tibetan-regular-indented">{Tsadra_template.sub_title_SP}{source_metadata.get("subtitle", "")}</span></p>\n'
        authors = source_metadata.get("authors", [])
        front_page += f"{book_title_tag}{sub_title_tag}"
        for author in authors:
            front_page += f"{Tsadra_template.author_SP}{author}</span></p>\n"
        if credit_pg_path.is_file():
            front_page += f'{Tsadra_template.credit_page_SP}<img src="{self.opf_path}/assets/image/{credit_pg_name}" alt="credit image not found"/></span></p>\n'
        return front_page

    def add_page_break(self, prev_p_tag, body_text):
        """Add page break before title annotation type p tag

        Args:
            prev_p_tag (str): prev tag of the title type p tag
            body_text (str): body text of ebook

        Returns:
            str: body text with page break before title type p tag
        """
        if not self.is_title(prev_p_tag) and (
            len(self.get_p_text(prev_p_tag)) > 500 or self.has_yigchung(prev_p_tag)
        ):
            new_prev_p_tag = prev_p_tag.replace(
                '<p class="tibetan-commentary-non-indent">',
                '<p class="tibetan-commentary-non-indent1">',
            )
            new_prev_p_tag = new_prev_p_tag.replace(
                '<p class="tibetan-regular-indented">',
                '<p class="tibetan-regular-indented1">',
            )
        else:
            new_prev_p_tag = prev_p_tag
        return body_text.replace(prev_p_tag, new_prev_p_tag)

    def indentation_adjustment(self, body_text):
        """Add required indentation styles to body text's p tags

        Args:
            body_text (str): body of the ebook

        Returns:
            str: body text having p tags with expected indentation styles
        """
        p_tags = self.get_p_tags(body_text)

        prev_p_tag = p_tags[0]
        body_text = prev_p_tag
        for p_tag in p_tags[1:]:
            cur_p_tag = ""
            if self.is_annotated_p_tag(prev_p_tag):
                if not self.is_title(p_tag):
                    cur_p_tag = self.rm_indentation(p_tag)
                else:
                    body_text = self.add_page_break(prev_p_tag, body_text)
                    cur_p_tag = self.add_indentation(p_tag)
            else:
                if self.is_annotated_p_tag(p_tag):
                    if not self.is_title(p_tag):
                        cur_p_tag = self.rm_indentation(p_tag)
                    else:
                        body_text = self.add_page_break(prev_p_tag, body_text)
                        cur_p_tag = self.add_indentation(p_tag)
                else:
                    if (
                        len(self.get_p_text(prev_p_tag)) < 50
                        and len(self.get_p_text(p_tag)) > 150
                    ):
                        cur_p_tag = p_tag.replace(
                            "<p>", '<p class="tibetan-commentary-non-indent">'
                        )
                    else:
                        cur_p_tag = self.add_indentation(p_tag)
            body_text += cur_p_tag
            prev_p_tag = cur_p_tag
        return body_text

    def get_footnote_references(self, footnotes):
        """Generate footnote references using footnote layer

        Args:
            footnotes (dict): footnote annotation information

        Returns:
            str: footnote references p tags
        """
        footnote_references = ""
        p_tag = '<p class = "tibetan-commentary-non-indent">'
        for footnote_id, footnote in footnotes.items():
            footnote_references += f'{p_tag}<a href="#fm{footnote_id}">{Tsadra_template.footnote_reference_SP} id="fr{footnote_id}">{footnote["footnote_ref"]}</span></a></p>'
        return footnote_references

    def set_toc_level(self, toc_levels, serialized_html):
        """Return toc outline by checking presences of title type annotations and sabche

        Args:
            toc_levels (dict): toc level outline given by user
            serialized_html (str): final serialize html of the opf

        Returns:
            dict: toc outline created using existing title type annotations or sabche
        """
        new_toc_levels = {}
        walker = 1
        for level, annotation_name in toc_levels.items():
            annotation_xpath = Tsadra_template.toc_xpaths.get(annotation_name, "")
            if annotation_xpath and annotation_name in serialized_html:
                new_toc_levels[walker] = annotation_xpath
                walker += 1
        return new_toc_levels

    def get_new_metadata(self, soup, meta_data):
        """Update meta data of opf

        Args:
            soup (bs4-tag): beautiful soup tag of whole opf
            meta_data (bs4-tag): beautiful soup tag of metadata

        Returns:
            bs4-tag: new metadata
        """
        ibook_meta = soup.new_tag('meta')
        ibook_meta.attrs['property'] = "ibooks:specified-fonts"
        ibook_meta.append('true')
        meta_data.append(ibook_meta)
        return meta_data

    def get_new_manifest(self, soup, manifest):
        """Update manifest of opf

        Args:
            soup (bs4-tag): beautiful soup tag of whole opf
            manifest (bs4-tag): beautiful soup tag of manifest

        Returns:
            bs4-tag: new manifest
        """
        new_item = soup.new_tag('item')
        new_item.attrs['href'] = "font/MonlamUniOuChan2.ttf"
        new_item.attrs['id'] = "MonlamUniOuChan2.ttf"
        new_item.attrs['media-type'] = "application/x-font-ttf"
        manifest.append(new_item)
        return manifest

    def get_new_opf(self, opf):
        """Updating opf content of epub with ibook specifications

        Args:
            opf (str): opf content of epub produced by calibre

        Returns:
            str: new opf content
        """
        soup = BeautifulSoup(opf, "html.parser")
        soup.package.attrs['prefix'] = "ibooks: http://vocabulary.itunes.apple.com/rdf/ibooks/vocabulary-extensions-1.0/"
        old_meta = soup.find('metadata')
        old_manifest = soup.find('manifest')
        new_meta = self.get_new_metadata(soup, old_meta)
        new_manifest = self.get_new_manifest(soup, old_manifest)
        soup.metadata.replaceWith(new_meta)
        soup.manifest.replaceWith(new_manifest)
        return str(soup)

    def embed_ibook_specific_font(self, epub_path):
        """Include ibook specification for proper font embedding and recompiling the ebook.

        Args:
            epub_path (path_obj): path of epub produced by calibre
        """
        pecha_id = epub_path.stem
        #rename to zip
        zip_path = epub_path.parent / f'{pecha_id}.zip'
        os.rename(str(epub_path), str(zip_path))
        # epub_path.rename(epub_path.with_suffix('.zip'))
        epub_folder = str(epub_path.parent / f'{pecha_id}')
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(epub_folder)
        print('INFO: Unzip epub..')
        zip_path.unlink()
        opf_path = Path(epub_folder) / 'content.opf'
        opf_content = opf_path.read_text(encoding='utf-8')
        new_opf_content = self.get_new_opf(opf_content)
        opf_path.write_text(new_opf_content, encoding='utf-8')
        print('INFO: OPF content updated..')
        new_zip_path = shutil.make_archive(epub_folder, 'zip', epub_folder)
        os.system(f'rm -r {epub_folder}')
        print('INFO: file zipped..')
        new_zip_path = Path(new_zip_path)
        new_epub_path = f'{epub_folder}.epub'
        os.rename(str(new_zip_path), new_epub_path)
        # new_zip_path.rename(new_zip_path.with_suffix('.epub'))
        print('INFO: Epub ready...')

    def get_serialized_html(self, result, base_id, pecha_title):
        """Serialize html are return using annotated text

        Args:
            result (str): tsadra epub annotations applied text
            base_id (str): volume id
            pecha_title (str): pecha title

        Returns:
            str: annotated applied text in html
        """
        result = f"{self.get_front_page()}{result}"
        footnote_ref_tag = ""
        if "Footnote" in self.layers:
            footnote_fn = self.opf_path / "layers" / base_id / "Footnote.yml"
            footnote_layer = load_yaml(footnote_fn)
            footnote_ref_tag = self.get_footnote_references(
                footnote_layer["annotations"]
            )
        result = self.p_tag_adder(result)
        result = self.indentation_adjustment(result)
        serialized_html = (
            f"<html>\n<head>\n\t<title>{pecha_title}</title>\n</head>\n<body>\n"
        )
        serialized_html += f"{result}{footnote_ref_tag}</body>\n</html>"
        return serialized_html

    def serialize(self,  output_path="./output/epub_output", toc_levels={}) -> Path:
        """
        This module serialize .opf file to other format such as .epub etc. In case of epub,
        we are using calibre ebook-convert command to do the conversion by passing our custom css template
        and embedding our custom font. The converted output will be then saved in current directory
        as {pecha_id}.epub.

        Args:
            toc_levels (dict): table of content level
            output_path (str): output path where epub needs to be saved

        Returns:
            str: output path

        """
        output_path = Path(output_path)
        out_html_fn = f"{self.meta['id']}.html"
        pecha_title = self.meta["source_metadata"].get("title", "")
        cover_image = self.meta["source_metadata"].get("cover", "")

        self.apply_layers()
        self.layers = [layer for layer in self.layers if layer != "Pagination"]

        results = self.get_result()
        for base_id, result in results.items():
            serialized_html = self.get_serialized_html(result, base_id, pecha_title)
            Path(out_html_fn).write_text(serialized_html)
            # Downloading css template file from ebook template repo and saving it
            template = requests.get(
                "https://raw.githubusercontent.com/OpenPecha/ebook-template/master/tsadra_template.css"
            )
            Path("template.css").write_bytes(template.content)
            # Running ebook-convert command to convert html file to .epub (From calibre)
            # XPath expression to detect chapter titles.
            if not toc_levels:
                toc_levels = Tsadra_template.toc_levels
            toc_levels = self.set_toc_level(toc_levels, serialized_html)
            level1_toc_Xpath = toc_levels.get(1, "")
            level2_toc_Xpath = toc_levels.get(2, "")
            level3_toc_Xpath = toc_levels.get(3, "")

            cover_path = self.opf_path / f"assets/image/{cover_image}"
            out_epub_fn = output_path / f"{self.meta['id']}.epub"
            font_family = "Monlam Uni Ouchan2"
            if cover_path.is_file():
                os.system(
                    f'ebook-convert {out_html_fn} {out_epub_fn} --extra-css=./template.css --embed-font-family="{font_family}" --page-breaks-before="{Tsadra_template.book_title_Xpath}" --cover={cover_path} --flow-size=0 --level1-toc="{level1_toc_Xpath}" --level2-toc="{level2_toc_Xpath}" --level3-toc="{level3_toc_Xpath}" --use-auto-toc --disable-font-rescaling'
                )
            else:
                os.system(
                    f'ebook-convert {out_html_fn} {out_epub_fn} --extra-css=./template.css --embed-font-family="{font_family}" --page-breaks-before="{Tsadra_template.book_title_Xpath}" --flow-size=0 --level1-toc="{level1_toc_Xpath}" --level2-toc="{level2_toc_Xpath}" --level3-toc="{level3_toc_Xpath}" --use-auto-toc --disable-font-rescaling'
                )
            # Removing html file and template file
            os.system(f"rm {out_html_fn}")
            os.system("rm template.css")
            if out_epub_fn.is_file():
                self.embed_ibook_specific_font(out_epub_fn)
            return out_epub_fn



from pathlib import Path

from bs4 import BeautifulSoup
from docx import Document
from docx.enum.style import WD_STYLE_TYPE
from docx.enum.text import WD_BREAK, WD_PARAGRAPH_ALIGNMENT
from docx.shared import Pt, RGBColor

from openpecha.serializers.epub import EpubSerializer
from openpecha.serializers.serialize import Serialize


class DocxSerializer(Serialize):

    def get_all_p_tags(self, html):
        """list all the p tags in serialized html

        Args:
            html (string): serialized html

        Returns:
            list: list of para tags in serialized html
        """
        soup = BeautifulSoup(html, "html.parser")
        p_tags = soup.find_all("p")
        return p_tags

    def format_span(self, span, para_doc_obj, document):
        """Add text and its custom style of span tag to document

        Args:
            span (bs4 obj): span tag
            para_doc_obj (para obj): para obj of document
            document (obj): document object in which text with its styles are accumilated
        """
        if span['class'][0] == "tibetan-root-text-in-verse":
            para_doc_obj.style = document.styles['Tsawa_verse_first_line']
            Tsawa = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-root-text-in-verse-middle-line" == span['class'][0]:
            para_doc_obj.style = document.styles['Tsawa_verse_middle_line']
            Tsawa = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-root-text-in-verse-last-line" == span['class'][0]:
            para_doc_obj.style = document.styles['Tsawa_verse_last_line']
            Tsawa = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-root-text" == span['class'][0]:
            Tsawa = para_doc_obj.add_run(f'{span.text}')
            Tsawa.style = document.styles['Tsawa_inline']

        elif span['class'][0] == "tibetan-citations-in-verse":
            para_doc_obj.style = document.styles['Citation_verse_first_line']
            Citation = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-citations-in-verse-middle-line" == span['class'][0]:
            para_doc_obj.style = document.styles['Citation_verse_middle_line']
            Citation = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-citations-in-verse-last-line" == span['class'][0]:
            para_doc_obj.style = document.styles['Citation_verse_last_line']
            Citation = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-citations-prose" == span['class'][0]:
            para_doc_obj.style = document.styles['Citation_prose']
            Citation = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-citations" == span['class'][0]:
            Citation = para_doc_obj.add_run(f'{span.text}')
            Citation.style = document.styles['Citation_inline']

        elif "tibetan-sabche" == span['class'][0]:
            para_doc_obj.style = document.styles['Heading 2']
            Sabche = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-sabche1" == span['class'][0]:
            Inline_sabche = para_doc_obj.add_run(f'{span.text}')
            Inline_sabche.style = document.styles['Inline_sabche']

        elif "tibetan-chapters" in span['class'][0]:
            para_doc_obj.style = document.styles['Heading 1']
            Chapter = para_doc_obj.add_run(f'{span.text}')

        elif "credits-page_front-title" in span['class'][0]:
            para_doc_obj.style = document.styles['Book_title']
            Booktitle = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-book-sub-title" in span['class'][0]:
            para_doc_obj.style = document.styles['Subtitle']
            Subtitle = para_doc_obj.add_run(f'{span.text}')

        elif "front-page---text-titles" in span['class'][0]:
            para_doc_obj.style = document.styles['Author']
            Author = para_doc_obj.add_run(f'{span.text}')

        elif "tibetan-commentary-small" == span['class'][0]:
            Yigchung = para_doc_obj.add_run(f'{span.text}')
            Yigchung.style = document.styles['Yigchung']

        else:
            normal_text = para_doc_obj.add_run(f'{span.text}')
            normal_text.style = document.styles['Commentary']

    def add_credit_page(self, p_tag, para_doc_obj):
        """Serialise credit page iamge in document

        Args:
            p_tag (obj): para tag object of bs4
            para_doc_obj (obj): para obj of document
        """
        img_tag = p_tag.find_all('img')
        img_path = img_tag[0]['src']
        credit_page = para_doc_obj.add_run()
        credit_page.add_picture(img_path)

    def page_break_needed(self, p_tag):
        """Return true of p tag's text need a page break after it

        Args:
            p_tag (obj): para tag object of bs4

        Returns:
            Boolean: True if page break needed after wards else False
        """
        p_tag_class = p_tag['class'][0]
        if p_tag_class == 'tibetan-commentary-non-indent1' or p_tag_class == 'tibetan-regular-indented1':
            return True
        else:
            return False

    def format_p_tag(self, p_tag, para_doc_obj, document):
        """Format p tag text with its custom style in document

        Args:
            p_tag (obj): para tag object of bs4
            para_doc_obj (obj): para obj of document
            document (obj): document object
        """
        if 'tibetan-regular-indented' in p_tag['class'][0]:
                para_doc_obj.style = document.styles['Indented_commentary']
        for content in p_tag.contents:
            if 'span' == content.name:
                self.format_span(content, para_doc_obj, document)
            else:
                normal_text = para_doc_obj.add_run(f'{content}')
                normal_text.style = document.styles['Commentary']
                if self.page_break_needed(p_tag):
                    normal_text.add_break = WD_BREAK.PAGE

    def add_styles(self, doc_styles):
        """Add all the custom style in document's style

        Args:
            doc_styles (obj): document's styles
        """

        #Book Title
        book_title_style = doc_styles.add_style('Book_title', WD_STYLE_TYPE.PARAGRAPH)
        book_title_font = book_title_style.font
        book_title_font.color.rgb = RGBColor(0,0,0)
        book_title_font.name = "Monlam Uni OuChan2"
        book_title_font.bold = True
        book_title_font.size = Pt(20)
        title_para_format = book_title_style.paragraph_format
        title_para_format.alignment = WD_PARAGRAPH_ALIGNMENT.CENTER
        title_para_format.space_after = Pt(30)
        title_para_format.space_before = Pt(15)

        #Subtitle
        subtitle_style = doc_styles['Subtitle']
        subtitle_style.font.color.rgb = RGBColor(0,0,0)
        subtitle_style.font.italic = False
        subtitle_style.font.name = "Monlam Uni OuChan2"
        subtitle_para_format = subtitle_style.paragraph_format
        subtitle_para_format.alignment = WD_PARAGRAPH_ALIGNMENT.CENTER
        subtitle_para_format.space_after = Pt(14)
        subtitle_para_format.space_before = Pt(14)

        #Author
        author_style = doc_styles.add_style('Author', WD_STYLE_TYPE.PARAGRAPH)
        author_style.font.color.rgb = RGBColor(0,0,0)
        author_style.font.size = Pt(10)
        author_style.font.name = "Monlam Uni OuChan2"
        author_para_format = author_style.paragraph_format
        author_para_format.alignment = WD_PARAGRAPH_ALIGNMENT.CENTER
        author_para_format.space_after = Pt(18)

        #Chapter
        chapter_style = doc_styles['Heading 1']
        chapter_style.font.color.rgb = RGBColor(0,0,0)
        chapter_style.font.name = "Monlam Uni OuChan2"
        chapter_style.font.bold = False
        chapter_style.font.size = Pt(16)
        chapter_para_format = chapter_style.paragraph_format
        chapter_para_format.space_after = Pt(14)
        chapter_para_format.space_before = Pt(14)


        #Sabche
        sabche_style = doc_styles['Heading 2']
        sabche_style.font.color.rgb = RGBColor(0,94,127)
        sabche_style.font.bold = False
        sabche_style.font.size = Pt(14)
        sabche_style.font.name = "Monlam Uni OuChan2"
        sabche_para_format = sabche_style.paragraph_format
        sabche_para_format.space_after = Pt(23)
        sabche_para_format.space_before = Pt(5)

        sabche_inline_style = doc_styles.add_style('Inline_sabche', WD_STYLE_TYPE.CHARACTER)
        sabche_inline_style.font.color.rgb = RGBColor(0,94,127)
        sabche_inline_style.font.size = Pt(14)
        sabche_inline_style.font.name = "Monlam Uni OuChan2"

        #Tsawa
        tsawa_style = doc_styles.add_style('Tsawa_verse_first_line', WD_STYLE_TYPE.PARAGRAPH)
        tsawa_style.font.color.rgb = RGBColor(139,20,9)
        tsawa_style.font.size = Pt(14)
        tsawa_style.font.name = "Monlam Uni OuChan2"
        tsawa_para_format = tsawa_style.paragraph_format
        tsawa_para_format.left_indent = Pt(48)
        tsawa_para_format.space_after = Pt(0)
        tsawa_para_format.space_before = Pt(5)

        tsawa_style = doc_styles.add_style('Tsawa_verse_middle_line', WD_STYLE_TYPE.PARAGRAPH)
        tsawa_style.font.color.rgb = RGBColor(139,20,9)
        tsawa_style.font.size = Pt(14)
        tsawa_style.font.name = "Monlam Uni OuChan2"
        tsawa_para_format = tsawa_style.paragraph_format
        tsawa_para_format.left_indent = Pt(48)
        tsawa_para_format.space_after = Pt(0)
        tsawa_para_format.space_before = Pt(0)

        tsawa_style = doc_styles.add_style('Tsawa_verse_last_line', WD_STYLE_TYPE.PARAGRAPH)
        tsawa_style.font.color.rgb = RGBColor(139,20,9)
        tsawa_style.font.size = Pt(14)
        tsawa_style.font.name = "Monlam Uni OuChan2"
        tsawa_para_format = tsawa_style.paragraph_format
        tsawa_para_format.left_indent = Pt(48)
        tsawa_para_format.space_after = Pt(20)
        tsawa_para_format.space_before = Pt(0)

        tsawa_inline_style = doc_styles.add_style('Tsawa_inline', WD_STYLE_TYPE.CHARACTER)
        tsawa_inline_style.font.color.rgb = RGBColor(139,20,9)
        tsawa_inline_style.font.size = Pt(14)
        tsawa_inline_style.font.name = "Monlam Uni OuChan2"

        #Citation
        citation_style = doc_styles.add_style('Citation_verse_first_line', WD_STYLE_TYPE.PARAGRAPH)
        citation_style.font.color.rgb = RGBColor(137,115,53)
        citation_style.font.size = Pt(14)
        citation_style.font.name = "Monlam Uni OuChan2"
        citation_para_format = citation_style.paragraph_format
        citation_para_format.left_indent = Pt(48)
        citation_para_format.space_after = Pt(0)
        citation_para_format.space_before = Pt(5)

        citation_style = doc_styles.add_style('Citation_verse_middle_line', WD_STYLE_TYPE.PARAGRAPH)
        citation_style.font.color.rgb = RGBColor(137,115,53)
        citation_style.font.size = Pt(14)
        citation_style.font.name = "Monlam Uni OuChan2"
        citation_para_format = citation_style.paragraph_format
        citation_para_format.left_indent = Pt(48)
        citation_para_format.space_after = Pt(0)
        citation_para_format.space_before = Pt(0)

        citation_style = doc_styles.add_style('Citation_verse_last_line', WD_STYLE_TYPE.PARAGRAPH)
        citation_style.font.color.rgb = RGBColor(137,115,53)
        citation_style.font.size = Pt(14)
        citation_style.font.name = "Monlam Uni OuChan2"
        citation_para_format = citation_style.paragraph_format
        citation_para_format.left_indent = Pt(48)
        citation_para_format.space_after = Pt(20)
        citation_para_format.space_before = Pt(0)

        citation_style = doc_styles.add_style('Citation_prose', WD_STYLE_TYPE.PARAGRAPH)
        citation_style.font.color.rgb = RGBColor(137,115,53)
        citation_style.font.size = Pt(14)
        citation_style.font.name = "Monlam Uni OuChan2"
        citation_para_format = citation_style.paragraph_format
        citation_para_format.left_indent = Pt(48)
        citation_para_format.right_indent = Pt(48)
        citation_para_format.space_after = Pt(20)
        citation_para_format.space_before = Pt(20)

        citation_inline_style = doc_styles.add_style('Citation_inline', WD_STYLE_TYPE.CHARACTER)
        citation_inline_style.font.color.rgb = RGBColor(137,115,53)
        citation_inline_style.font.size = Pt(14)
        citation_inline_style.font.name = "Monlam Uni OuChan2"

        #Yigchung
        yigchung_style = doc_styles.add_style('Yigchung', WD_STYLE_TYPE.CHARACTER)
        yigchung_style.font.color.rgb = RGBColor(0,0,0)
        yigchung_style.font.size = Pt(10)
        yigchung_style.font.name = "Monlam Uni OuChan2"

        #Commentary
        commentary_indent1_style = doc_styles.add_style('Indented_commentary', WD_STYLE_TYPE.PARAGRAPH)
        commentary_indent1_style.font.color.rgb = RGBColor(0,0,0)
        commentary_indent1_style.font.size = Pt(14)
        commentary_indent1_style.font.name = "Monlam Uni OuChan2"
        commentary_indent1_para_format = commentary_indent1_style.paragraph_format
        commentary_indent1_para_format.first_line_indent = Pt(18)

        commentary_style = doc_styles.add_style('Commentary', WD_STYLE_TYPE.CHARACTER)
        commentary_style.font.color.rgb = RGBColor(0,0,0)
        commentary_style.font.size = Pt(14)
        commentary_style.font.name = "Monlam Uni OuChan2"

    def create_docx(self, html, output_path, pecha_id):
        """Serialize html to docx file with custom style for annotated tags

        Args:
            html (string): serialized html using epubserializer
            output_path (obj): output path where docx file need to be saved
            pecha_id (string): pecha id
        """
        p_tags = self.get_all_p_tags(html)
        document = Document()
        doc_styles = document.styles
        self.add_styles(doc_styles)
        for p_tag in p_tags:
            para_doc_obj = document.add_paragraph()
            if p_tag.attrs and p_tag['class'][0] == 'credits-page_epub-edition-line':
                self.add_credit_page(p_tag, para_doc_obj)
            else:
                self.format_p_tag(p_tag, para_doc_obj, document)
        out_fn = output_path / f"{pecha_id}.docx"
        document.save(str(out_fn))
        return out_fn

    def serialize(self, output_path: Path, **kwargs) -> Path:
        output_path = Path(output_path)
        epub_serializer = EpubSerializer(self.opf_path)
        epub_serializer.apply_layers()
        pecha_title = epub_serializer.meta["source_metadata"].get("title", "")
        pecha_id = epub_serializer.meta["id"]
        results = epub_serializer.get_result()
        for base_id, result in results.items():
            serialized_html = epub_serializer.get_serialized_html(result, base_id, pecha_title)
            out_fn = self.create_docx(serialized_html, output_path, pecha_id)
            return out_fn



"""This module is to create pecha opf and alignment opa
Avialable functions:
    create_alignment_from_source_text: It will create pecha opf and alignment opa from a source text.
    create_alignment_from_tmx: It will create source and target opf and alignment opa from tmx path.
"""

import os
from pathlib import Path
from uuid import uuid4

from openpecha import config, github_utils
from openpecha.core.annotation import AnnBase, Span
from openpecha.core.layer import Layer, LayerEnum
from openpecha.core.pecha import OpenPechaFS
from openpecha.utils import dump_yaml, load_yaml

from ..parsers.tmx import parse_tmx
from ..segmenters.sentence import get_sentence_segments
from . import TMXAlignment


def get_segment_annotation(segment, char_walker):
    seg_len = len(segment) - 1
    segment_annotation = {
        uuid4().hex: AnnBase(span=Span(start=char_walker, end=char_walker + seg_len))
    }
    return segment_annotation, (char_walker + seg_len)


def get_segment_layer(text):
    segment_annotations = {}
    char_walker = 0
    segments = text.splitlines()
    for segment in segments:
        if segment:
            segment_annotation, end = get_segment_annotation(segment, char_walker)
            segment_annotations.update(segment_annotation)
        char_walker = end + 2

    segment_layer = Layer(
        annotation_type=LayerEnum.segment, annotations=segment_annotations
    )
    return segment_layer


def get_base_text(text):
    final_base = ""
    segments = text.splitlines()
    for segment in segments:
        final_base += segment + "\n"
    return final_base


def create_opf_repo(segmented_text, opf_path, title, lang, origin_type):
    pecha_id = opf_path.stem
    opf = OpenPechaFS(path=opf_path)
    layers = {"0001": {LayerEnum.segment: get_segment_layer(segmented_text)}}
    base_text = get_base_text(segmented_text)
    bases = {"0001": base_text}
    metadata = {
        "id": pecha_id,
        "initial_creation_type": "TMX",
        "origin_type": origin_type,
        "source_metadata": {
            "id": "",
            "title": title,
            "language": lang,
            "author": "",
            "volume": {},
        },
    }
    opf.layers = layers
    opf.bases = bases
    opf.save_base()
    opf.save_layers()
    meta_fn = Path(f"{opf_path}/meta.yml")
    dump_yaml(metadata, meta_fn)


def create_readme(pecha_path):
    pecha_id = pecha_path.stem
    meta_yml = load_yaml((pecha_path / f"{pecha_id}.opf/meta.yml"))
    pecha = f"|Pecha id | {pecha_id}"
    Table = "| --- | --- "
    Title = f"|Title | {meta_yml.get('source_metadata', {}).get('title', None)} "
    type = f"|Type | {meta_yml.get('origin_type', None)}"
    lang = f"|Language | {meta_yml.get('source_metadata',{}).get('language', None)}"
    Creator = f"|Initial creation | { meta_yml.get('initial_creation_type', None)}"
    readme = f"{pecha}\n{Table}\n{Title}\n{lang}\n{type}\n{Creator}"
    return readme


def create_opf(text, title=None, lang=None, origin_type="translation", new=False):
    if text:
        if new:
            text = get_sentence_segments(text)
        pecha_id = uuid4().hex
        opf_path = config.PECHAS_PATH / pecha_id / f"{pecha_id}.opf"
        create_opf_repo(text, opf_path, title, lang, origin_type)
        readme = create_readme(opf_path.parent)
        (opf_path.parent / "readme.md").write_text(readme, encoding="utf-8")
    return opf_path.parent


def publish_pecha(pecha_path):
    github_utils.github_publish(
        pecha_path,
        message="initial commit",
        not_includes=[],
        layers=[],
        org=os.environ.get("OPENPECHA_DATA_GITHUB_ORG"),
        token=os.environ.get("GITHUB_TOKEN"),
    )


def create_opf_from_tmx(tmx_path):
    title = tmx_path.stem

    src_text, tar_text, source_metadata = parse_tmx(tmx_path)
    if source_metadata["creationtool"] == "84000-translation-memory":
        origin_type = "translation"
    elif source_metadata["creationtool"] == "InterText":
        origin_type = "translation"
    else:
        origin_type = None

    src_lang = source_metadata.get("srclang", "")
    tar_lang = source_metadata.get("adminlang", "")

    source_pecha_path = create_opf(src_text, title, src_lang, origin_type)
    target_pecha_path = create_opf(tar_text, title, tar_lang, origin_type)

    return source_pecha_path, target_pecha_path, source_metadata


def create_alignment_from_source_text(text_path, title, publish=True):
    text = Path(text_path).read_text(encoding="utf-8")
    opf_path = create_opf(text, title=None, lang=None, new=True)
    obj = TMXAlignment()
    alignment_path = obj.create_alignment_repo(
        source_pecha_path=opf_path,
        target_pecha_path=None,
        title=title,
        source_metadata=None,
    )
    if publish:
        publish_pecha(opf_path)
        publish_pecha(alignment_path)
    return alignment_path


def create_alignment(source_pecha_path, target_pecha_path, title, source_metadata):
    obj = TMXAlignment()
    alignment_path = obj.create_alignment_repo(
        source_pecha_path, target_pecha_path, title, source_metadata
    )
    return alignment_path


def create_alignment_from_tmx(tmx_path, publish=True):
    title = tmx_path.stem
    source_pecha_path, target_pecha_path, source_metadata = create_opf_from_tmx(
        tmx_path
    )
    alignment_path = create_alignment(
        source_pecha_path, target_pecha_path, title, source_metadata
    )

    if publish:
        publish_pecha(source_pecha_path)
        publish_pecha(target_pecha_path)
        publish_pecha(alignment_path)

    return alignment_path



"""Module to support transifex integration.

Adding TM to transifex project is a premium feature.
This module provides hacky way to add TM to transifex project
by converting OpenPecha Alignment to transifex project resource
files.

Available classes:
- TransifexProject: class to represent transifex project
- OPATransifexProject: class to represent OpenPecha Alignment
    as transifex project

"""

import json
import logging
import os
import time
from pathlib import Path
from typing import List
from urllib.parse import urljoin

import requests
from slugify import slugify
from transifex.api import transifex_api
from transifex.api.jsonapi.exceptions import DoesNotExist

from .. import Alignment

API_TOKEN = os.getenv("TX_API_TOKEN")
transifex_api.setup(auth=API_TOKEN)


class TransifexProject:
    """Class representing translation project on transifex.

    Public methods:
    - add_resource: create resource with source file.
    - add_tm: create resource with source file and it's translations,
              which act as TM.
    """

    def __init__(self, org_slug: str, project_slug: str):
        self.org_slug = org_slug
        self.project_slug = project_slug
        org = transifex_api.Organization.get(slug=org_slug)
        self.project = org.fetch("projects").get(slug=project_slug)
        self.resources_api_url = "https://rest.api.transifex.com/resources"

    @property
    def id(self):
        return self.project.id

    @classmethod
    def create(
        cls,
        org_slug: str,
        project_name: str,
        repo_url: str,
        source_lang="bo",
        target_lang="en",
    ):
        """Create new translation project.

        This classmethod create new project to the organization
        but returns same project if it's already existed.

        Args:
            org_slug: organization slug
            project_name: name of project to be display in transifex.
            repo_url: project github repo url as it's madatory for public project.
            source_lang: source language code, default is Tibetan
            target_lang: target language code, default is English

        Returns:
            An instance of TransifexProject
        """

        project_slug = slugify(project_name)
        org = transifex_api.Organization.get(slug=org_slug)

        try:
            # check if projects exists
            project = org.fetch("projects").get(slug=project_slug)
            return cls(org_slug, project_slug)
        except DoesNotExist:
            # create new project
            source_language = transifex_api.Language.get(f"l:{source_lang}")
            target_language = transifex_api.Language.get(f"l:{target_lang}")

            project = transifex_api.Project.create(
                name=project_name,
                slug=project_slug,
                private=False,
                organization=org,
                source_language=source_language,
                repository_url=repo_url,
            )
            if not project:
                print(f"[ERROR] Cloud not create project ({project_name})")
                return

            # then add target language
            project.add("languages", [target_language])

            return cls(org_slug, project_slug)

    @property
    def languages(self):
        return self.project.fetch("languages")

    def _create_empty_resource(self, name, slug):
        """creates empty resource if doesn't exists yet and return resource id"""

        headers = {
            "Authorization": f"Bearer {API_TOKEN}",
            "Content-type": "application/vnd.api+json",
        }
        payload = {
            "data": {
                "attributes": {
                    "accept_translations": True,
                    "categories": ["category_1", "category_2"],
                    "name": name,
                    "priority": "normal",
                    "slug": slug,
                },
                "relationships": {
                    "i18n_format": {"data": {"id": "PO", "type": "i18n_formats"}},
                    "project": {"data": {"id": self.project.id, "type": "projects"}},
                },
                "type": "resources",
            }
        }

        response = requests.post(
            self.resources_api_url, headers=headers, data=json.dumps(payload)
        )
        if response.status_code == 200:
            return response.json()["data"]["id"]

    def _get_resource_id(self, slug: str):
        headers = {"Authorization": f"Bearer {API_TOKEN}"}
        url = f"{self.resources_api_url}/{self.project.id}:r:{slug}"
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            return response.json()["data"]["id"]

    def _get_resource(self, name: str, slug: str):
        resource_id = self._create_empty_resource(name, slug)
        if not resource_id:
            resource_id = self._get_resource_id(slug)
        if not resource_id:
            print("[ERROR] couldn't create the resouce or get existing resource")
            return
        return resource_id

    def add_resource(self, name: str, slug: str, source_path: Path):
        """Add resource to transifex project

        Args:
            name: name of the resource to be displayed in transifex
            slug: slug of resource, used for locating the resource
            source_path: path to source .po file

        Returns:
            Id of the resource issued by transifex
        """

        print(f"[INFO] adding resource to project ({self.project.id})")

        if not source_path.is_file():
            print(f"[ERROR] FileNotFound: {source_path}")
            return

        resource_id = self._get_resource(name, slug)
        if not resource_id:
            return

        resource = self.project.fetch("resources").get(slug=slug)
        transifex_api.ResourceStringsAsyncUpload.upload(
            resource=resource, content=source_path.open()
        )
        print(f"[INFO] Resource created at {resource_id}")
        return resource_id

    def add_translation(self, resource_id: str, lang: str, content_path: Path):
        """add translation .po file to a resource.

        Create async translation upload job and check if translation
        has been uploaded every 5 seconds.

        Args:
            resource_id: Id of a resource where translation .op file to be added.
            lang: lange code of the translation.
            content_path: path to translation (or target) .op file

        Returns:
            Upload job id of transifex
        """

        url = "https://rest.api.transifex.com/resource_translations_async_uploads"
        headers = {"Authorization": f"Bearer {API_TOKEN}"}
        payload = {"language": f"l:{lang}", "resource": resource_id}
        files = {"content": content_path.open("rb")}
        response = requests.post(url, headers=headers, data=payload, files=files)

        if response.status_code != 202:
            print("[ERROR] response:", response.json())
            return

        # check if upload has completed after every 5 secs
        upload_id = response.json()["data"]["id"]
        upload_status_check_url = f"https://rest.api.transifex.com/resource_translations_async_uploads/{upload_id}"

        while True:
            response = requests.get(upload_status_check_url, headers=headers)
            if response.status_code == 200:
                return upload_id
            time.sleep(5)

    def remove_resource(self, slug: str):
        """remvoes resource from the project"""

        resource = self.project.fetch("resources").get(slug=slug)
        resource_id = resource.id
        resource.delete()
        print(f"[INFO] Resource ({resource_id}) deleted")

    def _review_tm_translation(self, resource_slug, lang):
        """set review to true for tm translation"""
        language = transifex_api.Language.get(code=lang)
        resource = self.project.fetch("resources").get(slug=resource_slug)
        translations = transifex_api.ResourceTranslation.filter(
            resource=resource, language=language
        )
        for translation in translations:
            translation.save(reviewd=True)

    def add_tm(
        self,
        source_path: Path,
        target_path: Path,
        target_lang="en",
        resource_name="Translation Memory",
        resource_slug="tm",
    ):
        """Add Translation Memory to the project.

        Basically create resource with 100% translated source and traget .po files,
        which acts as a translation memory for actual resoruce to be translated

        Args:
            source_path: path to source .po file.
            target_path: path to target (translated) .po files.
            traget_lang: lang code to translation.
            resource_name: TM resource name to be displayed in transifex
            resource_slug: slug to locate TM resource.
        """

        print(f"[INFO] adding TM to project ({self.project.id})")

        resource_id = self.add_resource(
            name=resource_name, slug=resource_slug, source_path=source_path
        )
        if not resource_id:
            print(
                f"[ERROR] failed to add translation memory to project {self.project.id}"
            )
            return
        upload_id = self.add_translation(
            resource_id=resource_id, lang=target_lang, content_path=target_path
        )
        if not upload_id:
            print("[ERROR] failed to upload translation")
            self.remove_resource(slug=resource_slug)

        # self._review_tm_translation(resource_slug=resource_slug, lang=target_lang)

        print(f"[INFO] Successfully added TM to project ({self.project.id})")


class OPATransifexProject:
    """Represent OpenPecha Alignment as transifex project"""

    def __init__(self, org_slug: str, alignment_path: Path):
        self.org_slug = org_slug
        self.alignment = Alignment(path=alignment_path)
        self.tx_project = None
        self.github_org_url = "https://github.com/OpenPecha/"

    def create(self):
        """Creates OpenPecha Alignment project on transifex"""

        alignment_repo_url = urljoin(self.github_org_url, self.alignment.id)
        self.tx_project = TransifexProject.create(
            org_slug=self.org_slug,
            project_name=self.alignment.title,
            repo_url=alignment_repo_url,
        )
        logging.info(f"Project ({self.tx_project.id})")

    def start_new_translation(self):
        """upload source file to alignment"""
        po_view = self.alignment.get_po_view()
        self.tx_project.add_resource(
            name=self.alignment.title,
            slug=slugify(self.alignment.title),
            source_path=Path(po_view["source"]["path"]),
        )

    def import_translation(self):
        po_view = self.alignment.get_po_view()
        resource_id = self.tx_project.add_resource(
            name=self.alignment.title,
            slug=slugify(self.alignment.title),
            source_path=Path(po_view["source"]["path"]),
        )
        self.tx_project.add_translation(
            resource_id=resource_id,
            lang="en",
            content_path=Path(po_view["target"]["path"]),
        )

    def add_tm(self, alignment_ids: List[str]):
        """add `alignments` to the project."""

        added_tm_paths = []
        for alignment_id in alignment_ids:
            alignment = Alignment(id=alignment_id)
            po_view = alignment.get_po_view()
            resource_name = f"TM:{alignment.title}"
            self.tx_project.add_tm(
                source_path=Path(po_view["source"]["path"]),
                target_path=Path(po_view["target"]["path"]),
                target_lang=Path(po_view["target"]["lang"]),
                resource_name=resource_name,
                resource_slug=slugify(resource_name),
            )
            added_tm_paths.append(alignment.alignment_repo_path)
        return added_tm_paths


if __name__ == "__main__":
    project = TransifexProject.create(
        org_slug="esukhia",
        project_name="test tenzin project",
        repo_url="https://github.com/Esukhia/wordbook-translation",
    )
    # project = TransifexProject("esukhia", "test-2612")
    # project.remove_resource(slug="tm")
    # project.remove_resource(slug="translatethis")
    print("[INFO] Adding source file")
    project.add_resource(
        name="Translate this",
        slug="translatethis",
        source_path=Path("./tests/data/alignment/test-bo.po"),
    )
    project.add_tm(
        source_path=Path("./tests/data/alignment/test-bo.po"),
        target_path=Path("./tests/data/alignment/test-en.po"),
    )



"""
TODO: will be replaced by ./annotations.py
"""

from typing import Dict, Optional

from pydantic import BaseModel, validator


class Span(BaseModel):
    start: int
    end: int

    @validator("*")
    def span_must_not_be_neg(cls, v):
        if v < 0:
            raise ValueError("span shouldn't be negative")
        return v

    @validator("end")
    def end_must_not_be_less_than_start(cls, v, values, **kwargs):
        if "start" in values and v < values["start"]:
            raise ValueError("Span end must not be less than start")
        return v


class AnnBase(BaseModel):
    span: Span
    metadata: Optional[Dict] = None


class Page(AnnBase):
    page_info: Optional[str] = None  # page payload
    imgnum: Optional[
        int
    ] = None  # image sequence no. from bdrc api, http://iiifpres.bdrc.io/il/v:bdr:I0888
    reference: Optional[str] = None  # image filename from bdrc
